"""
compute and display conditial returns on short front month VIX future strategy SPVXSPI conditioned on
different predictors: VIX and vols
"""

import pandas as pd
import numpy as np
import seaborn as sns
import matplotlib.pyplot as plt
import qis as qis
from qis import PerfStat
from bbg_fetch import fetch_field_timeseries_per_tickers

# define names for convenience
benchmark_name = 'SPX'
strategy_name = 'ShortVix'
predictor_name = 'VIX'
benchmark_vol_name = f"{benchmark_name} real vol"
vix_vol_spread_name = f"Spread = {predictor_name} - real vol"
strategy_vol_name = f"{strategy_name} real vol"

# get index from bloomberg (run with open bloomberg terminal)
assets = {'SPXT Index': benchmark_name, 'SPVXSPI Index': strategy_name, 'VIX Index': predictor_name}
prices = fetch_field_timeseries_per_tickers(tickers=list(assets.keys()), field='PX_LAST', CshAdjNormal=True).dropna()
prices = prices.rename(assets, axis=1)

# define explanatory vars on monthly frequency, vol is multiplied by 100.0
monthly_return = prices[strategy_name].asfreq('ME', method='ffill').pct_change()
predictor_1 = prices[predictor_name].asfreq('ME', method='ffill').shift(1)
monthly_benchmark_vol_1 = 100.0*np.sqrt(252)*prices[benchmark_name].pct_change().rolling(21).std().asfreq('ME', method='ffill').shift(1).rename(benchmark_vol_name)
vix_vol_spread = predictor_1.subtract(monthly_benchmark_vol_1).rename(vix_vol_spread_name)
strategy_vol = 100.0*np.sqrt(252)*prices[strategy_name].pct_change().rolling(21).std().asfreq('ME', method='ffill').shift(1).rename(strategy_vol_name)
monthly_df = pd.concat([monthly_return, predictor_1, monthly_benchmark_vol_1, vix_vol_spread, strategy_vol], axis=1).dropna()

with sns.axes_style("darkgrid"):
    fig1, axs = plt.subplots(2, 3, figsize=(18, 9))
    qis.plot_prices_with_dd(prices=prices[[benchmark_name, strategy_name]],
                            regime_benchmark=benchmark_name,
                            perf_stats_labels=[PerfStat.PA_RETURN, PerfStat.VOL, PerfStat.SHARPE_RF0, PerfStat.MAX_DD],
                            title=f"Performances of ShortVix (SPVXSPI Index) and SPX (SPXT Index)",
                            x_date_freq='YE',
                            framealpha=0.9, fontsize=8,
                            axs=axs[:, 0])

    kwargs = dict(y=strategy_name,
                  num_buckets=6, ylabel=f"{strategy_name} monthly return",
                  yvar_format='{:.0%}', xvar_format='{:.0f}', showfliers=True, fontsize=10)
    qis.df_boxplot_by_classification_var(df=monthly_df,
                                         x=predictor_name,
                                         title=f"Monthly returns conditional on {predictor_name} at month start",
                                         x_hue_name=f"{predictor_name} month start",
                                         ax=axs[0, 1], **kwargs)

    qis.df_boxplot_by_classification_var(df=monthly_df,
                                         x=benchmark_vol_name,
                                         title=f"Monthly returns conditional on {benchmark_vol_name} at month start",
                                         x_hue_name=f"{benchmark_vol_name} month start",
                                         ax=axs[1, 1], **kwargs)

    qis.df_boxplot_by_classification_var(df=monthly_df,
                                         x=vix_vol_spread_name,
                                         title=f"Monthly returns conditional on {vix_vol_spread_name} at month start",
                                         x_hue_name=f"{vix_vol_spread_name} month start",
                                         ax=axs[0, 2], **kwargs)

    qis.df_boxplot_by_classification_var(df=monthly_df,
                                         x=strategy_vol_name,
                                         title=f"Monthly returns conditional on {strategy_vol_name} at month start",
                                         x_hue_name=f"{strategy_vol_name} month start",
                                         ax=axs[1, 2], **kwargs)
plt.show()



"""
common numpy operations
"""
import time
import numpy as np
import pandas as pd
from enum import Enum
from typing import Union, Optional, Tuple, Callable
from numba import njit


@njit
def np_apply_along_axis(func, axis: int, a: np.ndarray) -> np.ndarray:
    """
    numba function to apply func for 2-d array
    https://github.com/numba/numba/issues/1269
    """
    assert a.ndim == 2
    assert axis in [0, 1]

    if axis == 0:
        result = np.zeros((1, a.shape[1]))
        for i in range(a.shape[1]):
            result[0, i] = func(a[:, i])
    else:  # axis=1 with keep dims
        result = np.zeros((a.shape[0], 1))
        for i in range(a.shape[0]):
            result[i] = func(a[i, :])
    return result


@njit
def np_min(a: np.ndarray, axis: int = 1) -> np.ndarray:
    """
    min of 2-d array along axis
    """
    return np_apply_along_axis(func=np.min, axis=axis, a=a)


@njit
def np_nanmean(a: np.ndarray, axis: int = 1) -> np.ndarray:
    """
    nanmean of 2-d array along axis
    """
    return np_apply_along_axis(func=np.nanmean, axis=axis, a=a)


@njit
def np_nansum(a: np.ndarray, axis: int = 1, keep_dim: bool = True) -> np.ndarray:
    """
    nansum of 2-d array along axis
    """
    result = np_apply_along_axis(func=np.nansum, axis=axis, a=a)
    if not keep_dim:
        result = result.reshape(1, -1)
        result = result
    return result


@njit
def np_nanstd(a: np.ndarray, axis: int = 1, ddof: int = 1) -> np.ndarray:
    """
    nan std of 2-d array along axis
    """
    avg_std = np_apply_along_axis(func=np.nanstd, axis=axis, a=a)
    if ddof == 1:  # numbda does not recognise ddof as param to nanstd
        n = np_apply_along_axis(func=np.count_nonzero, axis=axis, a=~np.isnan(a))
        avg_std = np.where(n > 1, avg_std*np.sqrt(n / (n - 1.0)), np.nan)
    return avg_std


@njit
def np_nanvar(a: np.ndarray, axis: int = 1, ddof: int = 0) -> np.ndarray:
    """
    nan std of 2-d array along axis
    """
    avg_var = np_apply_along_axis(func=np.nanvar, axis=axis, a=a)
    if ddof == 1:  # numbda does not recognise ddof as param to nanstd
        n = np_apply_along_axis(func=np.count_nonzero, axis=axis, a=~np.isnan(a))
        avg_var = np.where(n > 1, avg_var*n / (n - 1.0), np.nan)
    return avg_var


@njit
def np_cumsum(a: np.ndarray, axis: int = 1) -> np.ndarray:
    """
    cannot do  return np_apply_along_axis(func=np.cumsum, axis=axis, a=a)
    because the axes is not reduced
    """
    result = np.zeros_like(a)
    if axis == 0:
        result[0] = a[0]
        for k in range(1, a.shape[0]):
            result[k] = result[k-1]+a[k]
    elif axis == 1:
        result[:, 0] = a[:, 0]
        if axis == 0:
            for k in range(1, a.shape[1]):
                result[:, k] = result[:, k - 1] + a[:, k]
    return result


@njit
def repeat_by_columns(a: np.ndarray, n: int) -> np.ndarray:
    return a.repeat(n).reshape((-1, n))


@njit
def repeat_by_rows(a: np.ndarray, n: int) -> np.ndarray:
    return a.repeat(n).reshape((-1, n)).T


@njit
def a_rank(a: np.ndarray, axis: int = 1) -> np.ndarray:
    order = a.argsort(axis=axis)  # get indices that sort array
    rank = order.argsort(axis=axis)
    return rank


@njit
def nan_func_to_data(a: np.ndarray,
                     func: Callable[[np.ndarray], np.ndarray] = np.nanmean,
                     axis: int = 0
                     ) -> np.ndarray:
    """
    apply nan sensitive function to data
    for pandas:  a.dim=2, a.shape = (nrows, ncols)
    row or columns wise operation equivalent to
    out = np.where(ind_all_nans, nans, np.nanmean(a=a, axis=axis))
    avoiding RuntimeWarning: Mean of empty slice
    """

    # simple case: a is one dimensional, axis do not matter then
    if a.ndim == 1:
        if axis == 0:
            if not np.all(np.isnan(a)):
                out = func(a)
            else:
                out = np.nan
        else:  # for 1-d array, we assume that func[a_n] = a_n
            raise ValueError(f"axis=1 not defined for 1-d array")

    else:
        nrows = a.shape[0]
        ncols = a.shape[1]

        if axis == 0:
            out = np.full(ncols, np.nan, dtype=np.double)
        else:
            out = np.full(nrows, np.nan, dtype=np.double)

        if axis == 0:  # column wise
            for col_idx in np.arange(ncols):
                col_a = a[:, col_idx]
                if not np.all(np.isnan(col_a)):
                    out[col_idx] = func(col_a)

        else:  # row wise
            for row_idx in np.arange(nrows):
                row_a = a[row_idx]
                if not np.all(np.isnan(row_a)):
                    out[row_idx] = func(row_a)

    return out


@njit
def to_finite(a: np.ndarray, fill_value: float = 0.0) -> np.ndarray:
    return np.where(np.isfinite(a), a, fill_value)


def to_finite_np(data: Union[pd.Series, pd.DataFrame, np.ndarray],
                 fill_value: float = 0.0,
                 a_min: Optional[float] = None,
                 a_max: Optional[float] = None,
                 is_min_max_clip_fill: bool = True  # will be filled at min or max
                 ) -> np.ndarray:
    """
    map to finite numbers
    note that pandas treans +/- inf differently than nan
    """
    if isinstance(data, pd.DataFrame) or isinstance(data, pd.Series):
        data_np = data.to_numpy()
    elif isinstance(data, np.ndarray):
        data_np = data.copy()
    else:
        raise TypeError(f"unsuported {type(data)}")

    data_np = to_finite(data_np, fill_value)

    if a_min is not None or a_max is not None:
        if is_min_max_clip_fill:
            data_np = np.clip(a=data_np, a_min=a_min, a_max=a_max)
        else:
            if a_max is not None:
                data_np = np.where(np.greater(data_np, a_max), np.nan, data_np)
            if a_min is not None:
                data_np = np.where(np.less(data_np, a_min), np.nan, data_np)

    return data_np


def to_finite_reciprocal(data: Union[pd.Series, pd.DataFrame, np.ndarray],
                         fill_value: float = 0.0,
                         is_gt_zero: bool = True,
                         a_min: float = None,
                         a_max: float = None
                         ) -> Union[pd.Series, pd.DataFrame, np.ndarray]:
    """
    finite repciprocal with inf fills
    """
    x_min, x_max = None, None
    if a_min is not None:
        x_max = 1.0 / a_min
    if a_max is not None:
        x_min = 1.0 / a_max

    x = to_finite_np(data, fill_value=fill_value, a_min=x_min, a_max=x_max)
    cond = None
    if is_gt_zero:
        cond = np.greater(x, 0.0)
    rec_x = np.where(cond, np.reciprocal(x, where=cond), np.nan)
    rec_x = to_finite(rec_x, fill_value)

    if isinstance(data, pd.DataFrame):
        rec_x = pd.DataFrame(rec_x, columns=data.columns, index=data.index)
    elif isinstance(data, pd.Series):
        rec_x = pd.Series(rec_x, name=data.name, index=data.index)

    return rec_x


def to_finite_ratio(x: Union[pd.Series, pd.DataFrame, np.ndarray],
                    y: Union[pd.Series, pd.DataFrame, np.ndarray],
                    fill_value: float = np.nan
                    ) -> Union[pd.Series, pd.DataFrame, np.ndarray]:
    """
    finite ratio x / y
    """
    x_np = to_finite_np(x, fill_value=fill_value)
    y_np = to_finite_np(y, fill_value=fill_value)
    x_y = np.divide(x_np, y_np, where=np.isclose(y_np, 0.0)==False)

    if isinstance(x, pd.DataFrame):
        x_y = pd.DataFrame(x_y, columns=x.columns, index=x.index)
    elif isinstance(x, pd.Series):
        x_y = pd.Series(x_y, name=x.name, index=x.index)

    return x_y


def covar_to_corr(covar: Union[np.ndarray, pd.DataFrame]) -> Union[np.ndarray, pd.DataFrame]:
    """
    compute correlation out of covariance
    """
    inv_vol = np.reciprocal(np.sqrt(np.diag(covar)))
    norm = np.outer(inv_vol, inv_vol)
    covar = covar.multiply(norm)
    return covar


def np_get_sorted_idx(a: np.ndarray) -> np.ndarray:
    """
    return indexes of sorted array
    """
    sorted_idx = a.argsort().argsort()  # need to apply double argsort
    return sorted_idx


def np_array_to_df_columns(a: np.ndarray, ncols: int) -> np.ndarray:
    """
    in case operation of np array is needed to apply to pandas column wise,
    the numpy array is broadcast to pandas columns
    the lenth = pandas.index
    """
    return np.tile(a, (ncols, 1)).T


def np_array_to_df_index(a: np.ndarray, n_index: int) -> np.ndarray:
    """
    in case operation of np array is needed to apply to pandas row wise,
    the numpy array is broadcast to pandas index
    the lenth = pandas.index
    """
    return np.tile(a, (n_index, 1))


def np_array_to_n_column_array(a: np.ndarray, ncols: int) -> np.ndarray:
    """
    for numpy broadcast of t-dim array to t-n dim array
    """
    return np.tile(a, (1, ncols))


def np_array_to_t_rows_array(a, t_rows: int) -> np.ndarray:
    """
    for numpy broadcast of t-dim array to t-n dim array
    """
    return np.tile(a, (t_rows, 1))


def np_array_to_matrix(a: np.ndarray, ncols: int) -> np.ndarray:
    """
    convert series too pandas dimension
    """
    matrix = np.broadcast_to(a, (ncols, len(a))).T
    return matrix


def np_matrix_add_array(matrix: np.ndarray,  # dim = T*N
                        array: np.ndarray,  # dim[a] = N or T(todo)
                        axis: int = 1  # 1 if dim[a] = N, 0 if dim[a] = T(todo)
                        ) -> np.ndarray:
    if axis == 1:
        m_array = np_array_to_t_rows_array(a=array, t_rows=matrix.shape[1])
        matrix_add = matrix + m_array
    else:
        raise ValueError(f"axis=0 is not imlemneted")
    return matrix_add


class RollFillType(Enum):
    NAN = 1  # empty is filled by nan
    ROLLOVER = 2  # corresponds to np.roll
    LAST_FILL = 3 # replace last left or right values


def np_shift(a: np.ndarray,
             shift: int,
             roll_fill_type: RollFillType = RollFillType.NAN
             ) -> np.ndarray:
    """
    shift numpy array
    """
    if a.ndim > 1:
        raise TypeError('only 1-d arrays are supported')
    n = a.shape[0]

    result = np.empty_like(a)

    if shift > 0:
        result[shift:] = a[:-shift]

        if roll_fill_type == RollFillType.NAN:
            result[:shift] = np.nan
        elif roll_fill_type == RollFillType.ROLLOVER:
            result[:shift] = a[n-shift:]
        elif roll_fill_type == RollFillType.LAST_FILL:
            result[:shift] = a[0]

    elif shift < 0:
        result[:shift] = a[-shift:]

        if roll_fill_type == RollFillType.NAN:
            result[shift:] = np.nan
        elif roll_fill_type == RollFillType.ROLLOVER:
            result[shift:] = a[:-shift]
        elif roll_fill_type == RollFillType.LAST_FILL:
            result[shift:] = a[-1]

    else:
        result[:] = a

    return result


@njit
def compute_expanding_power(n: int, power_lambda: float, reverse_columns: bool = False) -> np.ndarray:
    """
    compute expanding power = [1, lambda, lambda^2, ...]
    """
    a = np.log(power_lambda) * np.ones(n)
    a[0] = 0.0
    b = np.exp(np.cumsum(a))
    if reverse_columns:
        b = b[::-1]
    return b


def running_mean(x: np.ndarray, n: int, fill_value: float = np.nan) -> np.ndarray:
    x = to_finite_np(data=x, fill_value=fill_value)
    rolling_s = pd.Series(x).rolling(n, min_periods=0).apply(lambda x: np.nanmean(x))
    rolling_s = rolling_s.ffill()  # when x has sequence of nans longer than n
    return rolling_s.to_numpy()


def compute_paired_signs(x: np.ndarray,
                         y: np.ndarray,
                         ) -> Tuple[np.ndarray, np.ndarray, np.ndarray]:
    """
    compute signs equality of x and y excluding 0.0
    """
    joint_cond = np.logical_not(np.logical_and(np.isclose(x, 0.0), np.isclose(y, 0.0)))
    trend_cond = np.equal(np.sign(x), np.sign(y), where=joint_cond)
    rev_cond = np.not_equal(np.sign(x), np.sign(y), where=joint_cond)
    return joint_cond, trend_cond, rev_cond


@njit
def tensor_mean(a: np.ndarray) -> np.ndarray:
    """
    tensor mean in 1st direction
    typical application tensor of model params in t to compute the time average as (n1,n2)
    """
    nt = a.shape[0]
    n1 = a.shape[1]
    n2 = a.shape[2]
    output = np.zeros((n1, n2))
    for n1_ in range(n1):
        for n2_ in range(n2):
            output[n1_, n2_] = np.nanmean(a[:, n1_, n2_])
    return output


def find_nearest(a: np.ndarray,
                 value: float,
                 is_sorted: bool = True,
                 is_equal_or_largest: bool = False
                 ) -> float:
    """
    find closes element
    https://stackoverflow.com/questions/2566412/find-nearest-value-in-numpy-array
    """
    if is_sorted:
        idx = np.searchsorted(a, value, side="left")
        if is_equal_or_largest:  # return the equal or largest element
            return a[idx]
        else:
            if idx > 0 and (idx == len(a) or np.abs(value - a[idx - 1]) < np.abs(value - a[idx])):
                return a[idx - 1]
            else:
                return a[idx]
    else:
        a = np.asarray(a)
        idx = (np.abs(a - value)).argmin()
    return a[idx]


def to_nearest_values(a: np.ndarray,
                      values: np.ndarray,
                      is_sorted: bool = True
                      ) -> np.ndarray:
    """
    map array of values to nearest elements in array a
    """
    values_ = np.zeros_like(values)
    for idx, value in enumerate(values):
        values_[idx] = find_nearest(a=a, value=value, is_sorted=is_sorted)
    return values_


def compute_histogram_data(a: np.ndarray,
                           x_grid: np.ndarray,
                           name: str = 'Histogram'
                           ) -> pd.Series:
    """
    compute histogram on defined discrete grid
    """
    hist_data, bin_edges = np.histogram(a=a,
                                        bins=len(x_grid)-1,
                                        range=(x_grid[0], x_grid[-1]))
    hist_data = np.append(np.array(x_grid[0]), hist_data)
    hist_data = hist_data / len(a)
    hist_data = pd.Series(hist_data, index=bin_edges, name=name)
    return hist_data


def np_nonan_weighted_avg(a: np.ndarray, weights: np.ndarray) -> float:
    """
    compute weighted average of a using weights by masking nans in a
    """
    a = np.where(np.isfinite(a), a, np.nan)
    if np.all(np.isnan(a)):
        va = np.nan
    elif np.all(np.isclose(weights, 0.0)):
        va = 0.0
    else:
        ma = np.ma.MaskedArray(a, mask=np.isnan(a))
        va = np.ma.average(ma, weights=weights)
    return va


def set_nans_for_warmup_period(a: Union[np.ndarray, pd.DataFrame],
                               warmup_period: Union[int, np.ndarray]
                               ) -> Union[np.ndarray, pd.DataFrame]:
    """
    set nans for array a as warmup_period
    warmup_period starts from the first nonnan in a
    """
    if not (isinstance(warmup_period, int) or isinstance(warmup_period, np.ndarray)):
        raise ValueError(f"type={type(warmup_period)}")
    if isinstance(a, pd.DataFrame):
        a0 = a
        a = a.to_numpy()
    else:
        a0 = None
    if a.ndim == 2:
        if isinstance(warmup_period, int):
            warmup_period = np.full(a.shape[1], warmup_period)
        if np.any(warmup_period > 0):
            for idx, period in enumerate(warmup_period):
                nan_indicators = np.argwhere(np.isfinite(a[:, idx]))
                if nan_indicators.size > period:
                    a[:nan_indicators[period][0], idx] = np.nan
                else:
                    a[:, idx] = np.nan
    else:
        nan_indicators = np.argwhere(np.isfinite(a))
        if nan_indicators.size > warmup_period:
            a[:nan_indicators[warmup_period][0]] = np.nan
        else:
            a[:] = np.nan
    if a0 is not None:
        a = pd.DataFrame(a, index=a0.index, columns=a0.columns)

    return a


def select_non_nan_x_y(x: np.ndarray, y: np.ndarray) -> Tuple[np.ndarray, np.ndarray]:
    """
    select joint non-nan subset in x and y
    """
    assert x.shape[0] == y.shape[0]
    if y.ndim == 1:
        n_y = 1
    else:
        n_y = y.shape[1]

    if x.ndim == 1:
        n_x = 1
    else:
        n_x = x.shape[1]

    # ensure no nans
    if n_x == 1 and n_y == 1:
        is_nan_cond = np.logical_or(np.isnan(x), np.isnan(y))
        if np.any(is_nan_cond):
            x = x[is_nan_cond == False]
            y = y[is_nan_cond == False]
    elif n_x == 1 and n_y > 1:
        is_nan_cond = np.logical_or(np.isnan(x), np.isnan(y).any(axis=1))
        if np.any(is_nan_cond):
            x = x[is_nan_cond == False]
            y = y[is_nan_cond == False, :]
    elif n_x > 1 and n_y == 1:
        is_nan_cond = np.logical_or(np.isnan(x).any(axis=1), np.isnan(y))
        if np.any(is_nan_cond):
            x = x[is_nan_cond == False, :]
            y = y[is_nan_cond == False]
    else:
        is_nan_cond = np.logical_or(np.isnan(y).any(axis=1), np.isnan(x).any(axis=1))
        if np.any(is_nan_cond):
            x = x[is_nan_cond == False, :]
            y = y[is_nan_cond == False, :]
    return x, y


class LocalTests(Enum):
    SHIFT_TEST = 1
    CUM_POWER = 2
    ROLLING = 3
    WA = 4
    ARRAY_RANK = 5
    NEAREST = 6


def run_local_test(local_test: LocalTests):
    """Run local tests for development and debugging purposes.

    These are integration tests that download real data and generate reports.
    Use for quick verification during development.
    """

    if local_test == LocalTests.SHIFT_TEST:
        test_array = np.array([str(n) for n in range(10)])
        print('test_array')
        print(test_array)

        print('shift=2')
        for roll_fill_type in RollFillType:
            print(roll_fill_type)
            print(np_shift(a=test_array, shift=2, roll_fill_type=roll_fill_type))

        print('shift=-2')
        for roll_fill_type in RollFillType:
            print(roll_fill_type)
            print(np_shift(a=test_array, shift=-2, roll_fill_type=roll_fill_type))

    elif local_test == LocalTests.CUM_POWER:
        tic = time.perf_counter()
        for _ in np.arange(20):
            b = compute_expanding_power(n=10000000, power_lambda=0.97, reverse_columns=True)
        toc = time.perf_counter()
        print(f"{toc - tic} secs to run")
        print(b)

    elif local_test == LocalTests.ROLLING:
        x = np.array([1.0, 2.0, np.nan, np.nan, 3.0, 4.0, 5.0])
        xx = running_mean(x=x, n=2)
        print(xx)

    elif local_test == LocalTests.WA:
        x = np.array([1.0, 2.0, np.nan, np.nan, 3.0, 4.0, 5.0])
        weights = np.arange(len(x))
        print(x)
        print(weights)
        xx = np_nonan_weighted_avg(a=x, weights=weights)
        print(xx)

    elif local_test == LocalTests.ARRAY_RANK:
        a = np.array([1.0, 2.0, 10.0, 20.0, 3.0, 4.0, 5.0])
        array_rank = np.argsort(a).argsort()  # ranks by smallest value
        print(np.argsort(a))
        array_idx_rank = {array_rank[n]: n for n in np.arange(a.shape[0])}  # assign rank to idx
        array_idx_rank = dict(sorted(array_idx_rank.items()))  # sort by rank
        print(array_idx_rank)

    elif local_test == LocalTests.NEAREST:
        a = np.array([1.0, 2.0, 3.0, 4.0, 5.0])
        print(a)
        print(f"x={2.1}, nearest={find_nearest(a=a, value=2.1)}")
        print(f"x={2.1}, nearest={find_nearest(a=a, value=2.1, is_equal_or_largest=True)}")
        print(f"x={2.0}, nearest={find_nearest(a=a, value=2.0, is_equal_or_largest=True)}")


if __name__ == '__main__':

    run_local_test(local_test=LocalTests.SHIFT_TEST)



"""
module for computing rolling performance stats
"""
# packages
import numpy as np
import pandas as pd
import matplotlib.pyplot as plt
from scipy.stats import skew
from typing import Union, Optional, Tuple
from enum import Enum

# qis
import qis.perfstats.returns as ret
import qis.utils.annualisation
import qis.utils.dates as da
import qis.models.linear.ewm as ewm


class RollingPerfStat(Enum):
    """
    enumerated tuple for name and format
    """
    TOTAL_RETURNS = ('Total returns', '{:.2%}')
    PA_RETURNS = ('Pa returns', '{:.2%}')
    VOL = ('Volatility', '{:.2%}')
    SHARPE = ('Sharp ratio', '{:.2f}')
    SKEW = ('Skeweness', '{:.2f}')
    EWMA_VOL = ('EWMA vol', '{:.2%}')


def compute_rolling_perf_stat(prices: Union[pd.DataFrame, pd.Series],
                              rolling_perf_stat: RollingPerfStat = RollingPerfStat.TOTAL_RETURNS,
                              roll_freq: Optional[str] = 'B',
                              roll_periods: int = 260
                              ) -> Tuple[Union[pd.DataFrame, pd.Series], str]:
    """
    compute rolling performance
    default stats is 1y rolling for prices with roll_freq = 'B'
    for monthly returns anf 5y rolling use roll_freq='ME' and roll_periods=5*12
    """
    if roll_freq is None:
        roll_freq1 = pd.infer_freq(index=prices.index)
    else:
        roll_freq1 = roll_freq

    if rolling_perf_stat == RollingPerfStat.TOTAL_RETURNS:
        perf_stat = compute_rolling_returns(prices=prices, roll_freq=roll_freq, roll_periods=roll_periods)
        title = f"Rolling total returns with {roll_freq1}-freq and roll_period={roll_periods}"
    elif rolling_perf_stat == RollingPerfStat.PA_RETURNS:
        perf_stat = compute_rolling_pa_returns(prices=prices, roll_periods=roll_periods)
        title = f"P.a. returns with roll_period={roll_periods}"
    elif rolling_perf_stat == RollingPerfStat.VOL:
        perf_stat = compute_rolling_vols(prices=prices, roll_freq=roll_freq, roll_periods=roll_periods)
        title = f"Rolling vol with {roll_freq1}-freq and roll_period={roll_periods}"
    elif rolling_perf_stat == RollingPerfStat.SHARPE:
        perf_stat = compute_rolling_sharpes(prices=prices, roll_freq=roll_freq, roll_periods=roll_periods)
        title = f"Rolling Sharpe with {roll_freq1}-freq and roll_period={roll_periods}"
    elif rolling_perf_stat == RollingPerfStat.SKEW:
        perf_stat = compute_rolling_skew(prices=prices, roll_freq=roll_freq, roll_periods=roll_periods)
        title = f"Rolling skew with {roll_freq1}-freq and roll_period={roll_periods}"
    elif rolling_perf_stat == RollingPerfStat.EWMA_VOL:
        perf_stat = compute_ewma_vols(prices=prices, roll_freq=roll_freq, roll_periods=roll_periods)
        title = f"Rolling EWMA vol with {roll_freq1}-freq and span={roll_periods}"
    else:
        raise NotImplementedError(f"rolling_perf_stats")
    return perf_stat, title


def compute_rolling_returns(prices: Union[pd.DataFrame, pd.Series],
                            roll_freq: Optional[str] = None,
                            roll_periods: int = 260
                            ) -> Union[pd.DataFrame, pd.Series]:
    """
    compute rolling returns
    """
    if roll_freq is not None:
        prices = prices.asfreq(roll_freq, method='ffill')
    returns = prices.divide(prices.shift(periods=roll_periods)) - 1.0
    return returns


def compute_rolling_pa_returns(prices: Union[pd.DataFrame, pd.Series],
                               roll_periods: int = 260
                               ) -> Union[pd.DataFrame, pd.Series]:
    """
    compute rolling pa returns
    """
    pa_returns = prices.rolling(roll_periods).apply(lambda x: ret.compute_pa_return(x))
    return pa_returns


def compute_rolling_vols(prices: Union[pd.Series, pd.DataFrame],
                        roll_freq: Optional[str] = None,
                        roll_periods: int = 260
                        ) -> Union[pd.Series, pd.DataFrame]:
    log_returns = ret.to_returns(prices=prices, freq=roll_freq, is_log_returns=True, drop_first=False)
    saf = np.sqrt(qis.utils.annualisation.infer_an_from_data(data=log_returns))
    vols = saf * log_returns.rolling(roll_periods).apply(lambda x: np.nanstd(x, ddof=1))
    return vols


def compute_ewma_vols(prices: Union[pd.Series, pd.DataFrame],
                      roll_freq: Optional[str] = None,
                      roll_periods: int = 260
                      ) -> Union[pd.Series, pd.DataFrame]:
    log_returns = ret.to_returns(prices=prices, freq=roll_freq, is_log_returns=True, drop_first=False)
    vols = ewm.compute_ewm_vol(data=log_returns, span=roll_periods, annualize=True)
    return vols


def compute_rolling_sharpes(prices: Union[pd.Series, pd.DataFrame],
                            roll_freq: Optional[str] = None,
                            roll_periods: int = 260
                            ) -> Union[pd.Series, pd.DataFrame]:
    log_returns = ret.to_returns(prices=prices, freq=roll_freq, is_log_returns=True, drop_first=False)
    saf = np.sqrt(qis.utils.annualisation.infer_an_from_data(data=log_returns))
    sharpes = log_returns.rolling(roll_periods).apply(lambda x: compute_sharpe(x, saf=saf))
    return sharpes


def compute_sharpe(log_returns: Union[pd.Series, pd.DataFrame], saf: float = None) -> np.ndarray:
    if saf is None:
        saf = np.sqrt(qis.utils.annualisation.infer_an_from_data(data=log_returns))
    mean = np.expm1(np.nanmean(log_returns.to_numpy()))
    vol = np.nanstd(log_returns.to_numpy(), ddof=1)
    if np.greater(vol, 0.0):
        sharpe = saf * mean / vol
    else:
        sharpe = np.nan
    return sharpe


def compute_rolling_skew(prices: Union[pd.Series, pd.DataFrame],
                         roll_freq: Optional[str] = None,
                         roll_periods: int = 120,
                         ) -> Union[pd.Series, pd.DataFrame]:
    log_returns = ret.to_returns(prices=prices, freq=roll_freq, is_log_returns=True, drop_first=False)
    skw = log_returns.rolling(roll_periods).apply(lambda x: compute_skew(x))
    return skw


def compute_skew(log_returns: Union[pd.Series, pd.DataFrame]) -> np.ndarray:
    skw = skew(log_returns.to_numpy(), axis=0, nan_policy='omit')
    return skw


class LocalTests(Enum):
    ROLLING_STATS = 1


def run_local_test(local_test: LocalTests):
    """Run local tests for development and debugging purposes.

    These are integration tests that download real data and generate reports.
    Use for quick verification during development.
    """

    import qis.plots.time_series as pts
    from qis.test_data import load_etf_data
    prices = load_etf_data().dropna()

    if local_test == LocalTests.ROLLING_STATS:

        for rolling_perf_stat in RollingPerfStat:
            stats, title = compute_rolling_perf_stat(prices=prices,
                                              rolling_perf_stat=rolling_perf_stat,
                                              roll_freq='W-WED',
                                              roll_periods=5*52)
            pts.plot_time_series(df=stats,
                                 var_format='{:.2f}',
                                 title=f"{title}")

    plt.show()


if __name__ == '__main__':

    run_local_test(local_test=LocalTests.ROLLING_STATS)




# packages
import numpy as np
import pandas as pd
import matplotlib.pyplot as plt
import seaborn as sns
from typing import List, Union
from enum import Enum

# qis
import qis.file_utils as fu
import qis


# strats
from qis.portfolio.strats.quant_strats_delta1 import simulate_vol_target_strats_range, simulate_trend_strats_range

FIG_SIZE = (8.3, 11.7)  # A4 for portrait

PERF_PARAMS = qis.PerfParams(freq='B')


def create_time_series_report(prices: Union[pd.Series, pd.DataFrame],
                              time_period: qis.TimePeriod,
                              spans: List[int] = (7, 14, 21, 30, 60, 130, 260, 520),
                              vol_span: int = 33,
                              vol_target: float = 0.15,
                              vol_af: float = 260
                              ) -> List[plt.Figure]:

    if isinstance(prices, pd.Series):
        strat_name = prices.name
    else:
        strat_name = ''

    # vol target returns
    vt_nav_weights, vt_navs = simulate_vol_target_strats_range(prices=prices, vol_spans=spans,
                                                               vol_target=vol_target, vol_af=vol_af)

    fig1 = plt.figure(figsize=FIG_SIZE, constrained_layout=True)
    nrows = 5
    gs = fig1.add_gridspec(nrows=nrows, ncols=2, wspace=0.0, hspace=0.0)
    axs1 = [fig1.add_subplot(gs[row, 0]) for row in range(nrows)]
    axs2 = [fig1.add_subplot(gs[row, 1]) for row in range(nrows)]
    fig1.suptitle(f"{strat_name} Volatility-target strats with vol_target={vol_target:0.0%}", fontweight="bold", fontsize=8, color='blue')

    global_kwargs = dict(fontsize=5, linewidth=0.5, first_color_fixed=True, framealpha=0.8,
                         digits_to_show=1, x_date_freq='YE')
    scatter_kwargs = qis.update_kwargs(kwargs=global_kwargs,
                                       new_kwargs=dict(fontsize=5, var_format='{:.0%}', xvar_numticks=7,
                                                       markersize=1))

    plot_strategies_prices(nav_data=vt_navs,
                           nav_weights=vt_nav_weights,
                           time_period=time_period,
                           vol_span=vol_span,
                           vol_af=vol_af,
                           axs=axs1,
                           **global_kwargs)
    plot_strategies_returns_scatter(nav_data=vt_navs,
                                    time_period=time_period,
                                    axs=axs2,
                                    **scatter_kwargs)

    # tf returns
    tf_nav_weights, tf_navs, tf_signals = simulate_trend_strats_range(prices=prices, tf_spans=spans, vol_span=vol_span,
                                                                      vol_target=vol_target/2.0, vol_af=vol_af)

    fig2 = plt.figure(figsize=FIG_SIZE, constrained_layout=True)
    gs = fig2.add_gridspec(nrows=nrows, ncols=2, wspace=0.0, hspace=0.0)
    axs1 = [fig2.add_subplot(gs[row, 0]) for row in range(nrows)]
    axs2 = [fig2.add_subplot(gs[row, 1]) for row in range(nrows)]
    fig2.suptitle(f"{strat_name} Trend-following strats with vol_target={vol_target/2.0:0.0%}", fontweight="bold", fontsize=8, color='blue')

    plot_strategies_prices(nav_data=tf_navs,
                           nav_weights=tf_nav_weights,
                           time_period=time_period,
                           vol_span=vol_span,
                           vol_af=vol_af,
                           axs=axs1,
                           **global_kwargs)

    """
    figure_num += 1
    figure_caption = f"Figure {figure_num}. Pdf plot and qq-plot of returns on Trend-following Strategies by frequency"
    fig = plot_strategies_returns_pdf(nav_data=tf_navs,
                                      time_period=time_period,
                                      **global_kwargs)
    dut.add_fig_par(document, fig, figure_caption=figure_caption, figure_name=f"{file_name}_{figure_num}")
    """

    plot_strategies_returns_scatter(nav_data=tf_navs,
                                    time_period=time_period,
                                    axs=axs2,
                                    **scatter_kwargs)

    return [fig1, fig2]


def plot_strategies_prices(nav_data: pd.DataFrame,
                           nav_weights: pd.DataFrame,
                           time_period: qis.TimePeriod,
                           axs: List[plt.Subplot],
                           vol_span: int = 33,
                           vol_af: float = 260.0,
                           **kwargs
                           ) -> None:

    nav_returns = qis.to_returns(prices=nav_data)
    eod_ewm_vol = qis.compute_ewm_vol(data=nav_returns, span=vol_span, mean_adj_type=qis.MeanAdjType.NONE, annualization_factor=vol_af)

    # trim plot data
    nav_data = time_period.locate(nav_data)
    nav_weights = time_period.locate(nav_weights)
    nav_returns = time_period.locate(nav_returns)
    eod_ewm_vol = time_period.locate(eod_ewm_vol)

    qis.plot_prices(prices=nav_data,
                    var_format='{:.2f}',
                    title='Logarithm of nav',
                    perf_stats_labels=qis.PerfStatsLabels.DETAILED_WITH_SKEW.value,
                    is_log=True,
                    perf_params=PERF_PARAMS,
                    start_to_one=True,
                    ax=axs[0],
                    **kwargs)

    qis.plot_time_series(df=nav_weights,
                         var_format='{:.0%}',
                         title='Strategy weights',
                         legend_stats=qis.LegendStats.AVG_STD_LAST,
                         ax=axs[1],
                         **kwargs)

    turnover = nav_weights.diff(1).abs().rolling(int(vol_af)).sum()
    qis.plot_time_series(df=turnover,
                        var_format='{:.0%}',
                        title='1y rolling daily Turnover',
                        legend_stats=qis.LegendStats.AVG_STD_LAST,
                        ax=axs[2],
                        **kwargs)

    qis.plot_time_series(df=nav_returns,
                        var_format='{:.2%}',
                        title='Daily returns',
                        legend_stats=qis.LegendStats.AVG_STD_LAST,
                        ax=axs[3],
                        **kwargs)

    qis.plot_time_series(df=eod_ewm_vol,
                        var_format='{:.2%}',
                        title=f"Annualized EWMA-{vol_span} vol of daily returns",
                        legend_stats=qis.LegendStats.AVG_STD_LAST,
                        ax=axs[4],
                        **kwargs)


def plot_strategies_returns_pdf(nav_data: pd.DataFrame,
                                time_period: qis.TimePeriod,
                                **kwargs
                                ) -> plt.Figure:

    # trim plot data
    nav_data = time_period.locate(nav_data)

    freqs = {'Daily': 'B', 'Weekly': 'W-WED', 'Bi-Weekly': '2W-WED', 'Monthly': 'ME'}

    numticks = 7
    major_ticks = np.linspace(-4.5, 4.5, numticks)
    pdf_kwargs = qis.update_kwargs(kwargs=kwargs, new_kwargs={'fontsize': 7, 'var_format': '{:.0%}',
                                                              'xvar_numticks': numticks})

    qq_kwargs = qis.update_kwargs(kwargs=kwargs, new_kwargs={'fontsize': 7, 'var_format': '{:.1f}',
                                                             'yvar_numticks': numticks,
                                                             'xvar_major_ticks': major_ticks})

    with sns.axes_style("darkgrid"):
        fig, axs = plt.subplots(len(freqs.keys()), 2, figsize=FIG_SIZE, tight_layout=True)
        for idx, (title, freq) in enumerate(freqs.items()):
            returns = qis.to_returns(prices=nav_data, freq=freq, drop_first=True)
            qis.plot_histogram(df=returns,
                               x_min_max_quantiles=(0.0001, 0.9999),
                               title=f"{title}",
                               add_data_std_pdf=True,
                               ax=axs[idx][0],
                               **pdf_kwargs)
            qis.plot_qq(df=returns,
                    title=f"{title}",
                    ax=axs[idx][1],
                    **qq_kwargs)
    return fig


def plot_strategies_returns_scatter(nav_data: pd.DataFrame,
                                    time_period: qis.TimePeriod,
                                    axs: List[plt.Subplot],
                                    **kwargs
                                    ) -> None:
    # trim plot data
    nav_data = time_period.locate(nav_data)
    freqs = {'Daily': ('B', 252), 'Weekly': ('W-WED', 52), 'Bi-Weekly': ('2W-WED', 26), 'Monthly': ('ME', 12), 'Quarterly': ('QE', 4)}
    for ax, (title, freq) in zip(axs, freqs.items()):
        new_kwargs = dict(alpha_format='{0:+0.0%}',
                          beta_format='{:+0.1f}',
                          framealpha=0.9)
        kwargs = qis.update_kwargs(kwargs, new_kwargs)
        qis.plot_returns_scatter(prices=nav_data,
                                 benchmark=nav_data.columns[0],
                                 ylabel=f"Strategy return",
                                 title=f"{title}",
                                 freq=freq[0],
                                 order=2,
                                 ci=95,
                                 ax=ax,
                                 **kwargs)


class LocalTests(Enum):
    BTC_SIMULATION = 1
    SPY_SIMULATION = 2


def run_local_test(local_test: LocalTests):
    """Run local tests for development and debugging purposes.

    These are integration tests that download real data and generate reports.
    Use for quick verification during development.
    """

    import yfinance as yf

    if local_test == LocalTests.BTC_SIMULATION:
        prices = yf.download(tickers=['BTC-USD'], start="2003-12-31", end=None, ignore_tz=True, auto_adjust=True)['Close'].dropna()

        time_period = qis.TimePeriod('31Dec2015', '21Jun2023')
        figs = create_time_series_report(prices=prices, time_period=time_period, vol_target=0.5, vol_af=360)
        fu.save_figs_to_pdf(figs=figs, file_name='btc_analysis', orientation='landscape',
                            add_current_date=True, local_path=None)

    elif local_test == LocalTests.SPY_SIMULATION:
        prices = yf.download(tickers=['SPY'], start="2003-12-31", end=None, ignore_tz=True, auto_adjust=True)['Close'].dropna()
        time_period = qis.TimePeriod('31Dec2019', '15Aug2024')
        figs = create_time_series_report(prices=prices, time_period=time_period, vol_target=0.15, vol_af=260)
        fu.save_figs_to_pdf(figs=figs, file_name='spy_analysis', orientation='landscape',
                            add_current_date=True, local_path=None)

    plt.show()


if __name__ == '__main__':

    run_local_test(local_test=LocalTests.SPY_SIMULATION)



"""
reporting anatytics for smart diversification frontier of futures_strats and portfolios
"""
# packages
import warnings
import pandas as pd
import numpy as np
import matplotlib.pyplot as plt
import seaborn as sns
from dataclasses import dataclass
from enum import Enum
from typing import Dict, Union, List, Optional, Tuple
import qis as qis
from qis import PerfStat, RegimeData, BenchmarkReturnsQuantilesRegime, PerfParams

regime_classifier = BenchmarkReturnsQuantilesRegime(freq='QE')
PERF_PARAMS = PerfParams(freq='ME')


PERF_COLUMNS = (
    # PerfStat.START_DATE,
    # PerfStat.END_DATE,
    PerfStat.PA_RETURN,
    PerfStat.VOL,
    PerfStat.SHARPE_RF0,
    PerfStat.BEAR_SHARPE,
    PerfStat.NORMAL_SHARPE,
    PerfStat.BULL_SHARPE,
    PerfStat.MAX_DD,
    PerfStat.MAX_DD_VOL,
    PerfStat.SKEWNESS)


@dataclass
class SmartDiversificationReport:
    """
    class to build diversification fronters
    """
    overlay_navs: pd.DataFrame
    principal_nav: pd.Series = None
    regime_classifier: BenchmarkReturnsQuantilesRegime = None
    perf_params: PerfParams = None
    benchmark_description: str = None

    def __post_init__(self):
        if self.regime_classifier is None:
            self.regime_classifier = regime_classifier
        if self.perf_params is None:
            self.perf_params = PERF_PARAMS
        self.regime_classifier = BenchmarkReturnsQuantilesRegime()

    def compute_smart_diversification_curve(self,
                                            principal_nav: pd.Series,
                                            overlay_nav: pd.Series,
                                            principal_weight: float = 1.0,
                                            max_overlay_weight: float = 1.0,
                                            rebalancing_freq: str = 'QE',
                                            is_principal_weight_fixed: bool = True,
                                            x_var: PerfStat = PerfStat.BEAR_SHARPE,
                                            y_var: PerfStat = PerfStat.SHARPE_RF0
                                            ) -> pd.DataFrame:
        """
        compute diversification curve
        # the first portfolio is 100% of the benchmark
        """
        portfolio_navs = create_overlay_portfolio_curve(principal_nav=principal_nav,
                                                        overlay_nav=overlay_nav,
                                                        principal_weight=principal_weight,
                                                        max_overlay_weight=max_overlay_weight,
                                                        rebalancing_freq=rebalancing_freq,
                                                        is_principal_weight_fixed=is_principal_weight_fixed)
        cvar_table, _ = self.regime_classifier.compute_regimes_pa_perf_table(prices=portfolio_navs,
                                                                             benchmark=portfolio_navs.columns[0],
                                                                             perf_params=self.perf_params)
        return cvar_table[[x_var.to_str(), y_var.to_str()]]

    def get_overlay_points(self,
                           principal_nav: pd.Series,
                           x_var: PerfStat = PerfStat.BEAR_SHARPE,
                           y_var: PerfStat = PerfStat.SHARPE_RF0
                           ) -> pd.DataFrame:

        portfolio_navs = pd.concat([principal_nav, self.overlay_navs], axis=1)

        # the first 100% is the benchmark
        cvar_table, _ = self.regime_classifier.compute_regimes_pa_perf_table(prices=portfolio_navs,
                                                                             benchmark=portfolio_navs.columns[0],
                                                                             perf_params=self.perf_params)
        return cvar_table[[x_var.to_str(), y_var.to_str()]]

    def plot_nav(self,
                 var_format: str = '{:.0%}',
                 sharpe_format: str = '{:.2f}',
                 principal_nav: pd.Series = None,
                 is_include_principal: bool = False,
                 title: str = None,
                 is_log: bool = True,
                 ax: plt.Subplot = None,
                 **kwargs
                 ) -> plt.Figure:

        if principal_nav is None:
            principal_nav = self.principal_nav

        if is_include_principal:
            prices = pd.concat([principal_nav, self.overlay_navs], axis=1)
        else:
            prices = self.overlay_navs.copy()

        fig = qis.plot_prices(prices=prices,
                              perf_params=self.perf_params,
                              start_to_one=True,
                              is_log=is_log,
                              var_format=var_format,
                              sharpe_format=sharpe_format,
                              title=title,
                              ax=ax,
                              **kwargs)

        if ax is None:
            ax = fig.axes[0]
        """
        classification_data = sdi.create_regime_classification(regime_classifier=self.regime_classifier,
                                                               pivot_prices=principal_nav)

        qis.add_regime_shadows_to_ax(ax=ax,
                                     regime_classifier=self.regime_classifier,
                                     classification_data=classification_data,
                                     pivot_prices=principal_nav,
                                     price_data_index=principal_nav.index)
        """
        return fig

    def plot_ra_table(self,
                      time_period: qis.TimePeriod = None,
                      desc_map: Dict[str, str] = None,
                      ticker_map: Dict[str, str] = None,
                      perf_columns: List[PerfStat] = PERF_COLUMNS,
                      columns_title: str = 'Assets',
                      ax: plt.Subplot = None,
                      **kwargs
                      ) -> plt.Figure:

        prices = pd.concat([self.principal_nav, self.overlay_navs], axis=1)

        if time_period is not None:
            prices = time_period.locate(prices)

        benchmark = self.principal_nav.name if isinstance(self.principal_nav, pd.Series) else self.principal_nav.columns[0]
        cvar_table, _ = self.regime_classifier.compute_regimes_pa_perf_table(prices=prices,
                                                                             benchmark=benchmark,
                                                                             perf_params=self.perf_params)
        table_data = pd.DataFrame(data=prices.columns, index=cvar_table.index, columns=[columns_title])

        col_widths = [15]
        if desc_map is not None:
            table_data['Name'] = table_data.index.map(desc_map)
            col_widths.append(15)
        if ticker_map is not None:
            table_data['BBG\nTicker'] = table_data.index.map(ticker_map)
            col_widths.append(15)

        for perf_column in perf_columns:
            table_data[perf_column.to_str()] = qis.series_to_str(ds=cvar_table[perf_column.to_str()],
                                                             var_format=perf_column.to_format(**kwargs))
            col_widths.append(7)

        special_columns_colors = [(0, 'steelblue')]
        fig = qis.plot_df_table(df=table_data,
                                col_widths=col_widths,
                                column_width=7,
                                add_index_as_column=False,
                                index_column_name='Strategies',
                                special_columns_colors=special_columns_colors,
                                row_height=0.5,
                                ax=ax,
                                **kwargs)
        return fig

    def plot_conditional_sharpes(self,
                                 principal_nav: pd.Series = None,
                                 overlay_navs: pd.DataFrame = None,
                                 regime_data_to_plot: RegimeData = RegimeData.REGIME_SHARPE,
                                 var_format: str = '{:.2f}',
                                 is_names_to_2lines: bool = False,
                                 title: str = None,
                                 ax: plt.Subplot = None,
                                 **kwargs
                                 ) -> plt.Figure:

        if principal_nav is None:
            principal_nav = self.principal_nav

        if overlay_navs is None:
            overlay_navs = self.overlay_navs

        prices = pd.concat([principal_nav, overlay_navs], axis=1)
        benchmark = principal_nav.name if isinstance(principal_nav, pd.Series) else principal_nav.columns[0]

        if is_names_to_2lines:
            prices.columns = [x.replace(' ', '\n') for x in prices.columns]
            prices.columns = [x.replace('+', '\n+') for x in prices.columns]
            benchmark = benchmark.replace(' ', '\n')

        fig = qis.plot_regime_data(regime_classifier=self.regime_classifier,
                                   prices=prices,
                                   benchmark=benchmark,
                                   is_conditional_sharpe=True,
                                   regime_data_to_plot=regime_data_to_plot,
                                   var_format=var_format or '{:.2f}',
                                   perf_params=self.perf_params,
                                   title=title,
                                   ax=ax,
                                   **kwargs)
        return fig

    def plot_corr_matrix(self,
                         var_format: str = '{:.0%}',
                         freq: Optional[str] = None,
                         ax: plt.Subplot = None,
                         **kwargs
                         ) -> plt.Figure:
        if ax is None:  # create new axis
            height = qis.calc_table_height(num_rows=len(self.overlay_navs.columns),
                                                       first_row_height=2.0)
            fig, ax = plt.subplots(1, 1, figsize=(height, height))
        else:
            fig = None
        qis.plot_returns_corr_table(prices=self.overlay_navs,
                                    freq=freq,
                                    var_format=var_format,
                                    ax=ax,
                                    **kwargs)
        return fig

    def plot_smart_diversification_curve(self,
                                         principal_nav: pd.Series = None,
                                         principal_weight: float = 1.0,
                                         is_principal_weight_fixed: bool = True,
                                         rebalancing_freq: str = 'QE',
                                         x_var: PerfStat = PerfStat.BEAR_SHARPE,
                                         y_var: PerfStat = PerfStat.SHARPE_RF0,
                                         xlabel: str = None,
                                         ylabel: str = None,
                                         title: str = None,
                                         constraints: Dict[str, float] = None,
                                         shifts: Dict[str, bool] = None,
                                         shifts_n: Dict[str, float] = None,
                                         with_filter: List[str] = None,
                                         x_limits: Tuple[Optional[float], Optional[float]] = None,
                                         y_limits: Tuple[Optional[float], Optional[float]] = None,
                                         ax: plt.Subplot = None,
                                         **kwargs
                                         ) -> plt.Figure:

        if y_var.to_str() == x_var.to_str():
            raise ValueError('x_var and y_var cannot be the same')

        if principal_nav is None:
            principal_nav = self.principal_nav

        xy_datas = {}
        data_labels = []
        y_0 = None
        for idx, asset in enumerate(self.overlay_navs.columns):
            max_overlay_weight = 1.0
            if constraints is not None:
                if asset in constraints.keys():
                    max_overlay_weight = constraints[asset]
            name = asset
            if shifts_n is not None:
                if asset in shifts_n.keys():
                    name = asset + '\n'

            xy = self.compute_smart_diversification_curve(principal_nav=principal_nav,
                                                          overlay_nav=self.overlay_navs[asset],
                                                          principal_weight=principal_weight,
                                                          is_principal_weight_fixed=is_principal_weight_fixed,
                                                          max_overlay_weight=max_overlay_weight,
                                                          rebalancing_freq=rebalancing_freq,
                                                          x_var=x_var,
                                                          y_var=y_var)
            if with_filter is not None and asset in with_filter:
                y = xy[y_var.to_str()].to_numpy()
                x = xy[x_var.to_str()].to_numpy()
                x_lin = np.linspace(start=x[0], stop=x[-1], num=len(x)) # replace x with smooth linear grid
                poly_coeffs = safe_polyfit(x_lin, y, 3)
                y_smooth = np.poly1d(poly_coeffs)(x_lin)
                # Adjust to match endpoint
                if y_0 is None:
                    y_0 = y[0]
                y_smooth += np.linspace(0, y_0 - y_smooth[0], len(x_lin))
                y_smooth[0] = y_0

                xy[y_var.to_str()] = y_smooth
                xy[x_var.to_str()] = x_lin

            portfolio_labels = ['' for _ in xy.index]
            portfolio_labels[0] = f"{principal_nav.name}\n{'{:.0%}'.format(principal_weight)}"
            portfolio_labels[-1] = f"{name}  {'{:.0%}'.format(max_overlay_weight)}"
            if shifts is not None:
                if asset in shifts.keys():
                    if shifts[asset]:  # up
                        portfolio_labels[-1] = name + ' ' + '{:.0%}'.format(max_overlay_weight) + '\n'
                    else: #down
                        portfolio_labels[-1] = ' \n' + name + ' ' + '{:.0%}'.format(max_overlay_weight)

                if principal_nav.name in shifts.keys():
                    portfolio_labels[0] = f"{principal_nav.name} {'{:.0%}'.format(principal_weight)} \n"

            xy_datas[asset] = xy
            data_labels.append(portfolio_labels)

        markers = qis.get_n_markers(n=len(xy_datas.keys()))
        xlabel = xlabel or x_var.to_str()
        ylabel = ylabel or y_var.to_str()

        fig = qis.plot_lines_list(xy_datas=xy_datas,
                                  data_labels=data_labels,
                                  title=title,
                                  xlabel=xlabel,
                                  ylabel=ylabel,
                                  xvar_format=x_var.to_format(**kwargs),
                                  yvar_format=y_var.to_format(**kwargs),
                                  markers=markers,
                                  x_limits=x_limits,
                                  y_limits=y_limits,
                                  ax=ax,
                                  **kwargs)
        return fig

    def plot_smart_diversification_scatter(self,
                                           principal_nav: pd.Series = None,
                                           is_drop_principal: bool = False,
                                           x_var: PerfStat = PerfStat.BEAR_SHARPE,
                                           y_var: PerfStat = PerfStat.SHARPE_RF0,
                                           xvar_format: str = '{:.1f}',
                                           yvar_format: str = '{:.1f}',
                                           shifts: Dict[str, bool] = None,
                                           xlabel: Union[str, bool, None] = True,
                                           ylabel: Union[str, bool, None] = True,
                                           is_add_model_equation: bool = True,
                                           annotation_labels: List[str] = None,
                                           portfolio_curve: List[str] = None,
                                           ci: int = 95,
                                           ax: plt.Subplot = None,
                                           **kwargs
                                           ) -> plt.Figure:

        if principal_nav is None:
            principal_nav = self.principal_nav

        overlay_points = self.get_overlay_points(principal_nav=principal_nav,
                                                 x_var=x_var, y_var=y_var)
        if is_drop_principal:
            overlay_points = overlay_points.drop(index=principal_nav.name)

        if annotation_labels is None:
            if shifts is not None:
                annotation_labels = []
                for asset in overlay_points.index.to_list():
                    annotation_label = asset
                    if asset in shifts.keys():
                        if shifts[asset]:
                            annotation_label = ' \n' + asset + '\n '
                        else:
                            annotation_label = asset + ' \n'
                    annotation_labels.append(annotation_label)
            else:
                annotation_labels = overlay_points.index.to_list()

        fig = qis.plot_scatter(df=overlay_points,
                               x=x_var.to_str(),
                               y=y_var.to_str(),
                               xvar_format=xvar_format,
                               yvar_format=yvar_format,
                               ax=ax,
                               annotation_labels=annotation_labels,
                               add_universe_model_label=is_add_model_equation,
                               xlabel=xlabel,
                               ylabel=ylabel,
                               colors=['#00284A'],
                               full_sample_label='Cross-sectional fit: ',
                               ci=ci,
                               **kwargs)

        if portfolio_curve is not None:
            curve = overlay_points.loc[np.isin(overlay_points.index, portfolio_curve, assume_unique=True), :]
            y = curve[y_var.to_str()].to_numpy()
            x = curve[x_var.to_str()].to_numpy()
            try:
                curve_poly = np.polyfit(x, y, 2)  # fit 2-order polynomial to x and y curve
                x_lin = np.linspace(np.min(x), np.max(x), 50)  # make linear space between x_min and x_max
                y_curve = np.poly1d(curve_poly)(x_lin)  # fill in
                sns.lineplot(x=x_lin, y=y_curve, marker='None', color='blue', ax=ax)
            except np.linalg.LinAlgError:
                warnings.warn(f"SVD did not converge in Linear Least Squares")

        return fig

    def plot_returns_scatter(self,
                             principal_nav: pd.Series = None,
                             overlay_navs: pd.DataFrame = None,
                             add_45line: bool = False,
                             title: Union[str, None] = None,
                             ax: plt.Subplot = None,
                             **kwargs
                             ) -> plt.Figure:

        if principal_nav is None:
            principal_nav = self.principal_nav
        if overlay_navs is None:
            overlay_navs = self.overlay_navs
        prices = pd.concat([principal_nav, overlay_navs], axis=1)
        fig = qis.plot_returns_scatter(prices=prices,
                                       benchmark=str(principal_nav.name),
                                       add_45line=add_45line,
                                       #ylabel=f"{self.overlay_nav.name}",
                                       title=title,
                                       freq='QE',
                                       order=2,
                                       ci=95,
                                       ax=ax,
                                       **kwargs)
        return fig


def create_overlay_portfolio_curve(principal_nav: pd.Series,
                                   overlay_nav: pd.Series,
                                   principal_weight: float = 1.0,
                                   max_overlay_weight: float = 1.0,
                                   rebalancing_freq: str = 'QE',
                                   is_principal_weight_fixed: bool = True
                                   ) -> pd.DataFrame:
    """
    compute overlays for weights from 0% to 100%
    """
    prices = pd.concat([principal_nav, overlay_nav], axis=1)

    overlay_weights = np.linspace(0, max_overlay_weight, 11)
    portfolio_navs = []
    for overlay_weight in overlay_weights:
        if is_principal_weight_fixed:
            weights = np.array([principal_weight, overlay_weight])
        else:
            weights = np.array([1.0-overlay_weight, overlay_weight])

        portfolio_nav = qis.backtest_model_portfolio(prices=prices,
                                                 weights=np.array(weights),
                                                 rebalancing_freq=rebalancing_freq).get_portfolio_nav()
        portfolio_nav.name = f"{overlay_nav.name} {'{:.2%}'.format(overlay_weight)}"
        portfolio_navs.append(portfolio_nav)
    portfolio_navs = pd.concat(portfolio_navs, axis=1)
    return portfolio_navs


def safe_polyfit(x, y, degree=3) -> np.ndarray:
    """Robust polynomial fitting with fallback"""
    # Remove NaN values
    mask = ~(np.isnan(x) | np.isnan(y))
    x_clean = x[mask]
    y_clean = y[mask]

    if len(x_clean) < degree + 1:
        # Not enough points for requested degree
        degree = max(1, len(x_clean) - 1)
    """
    # Normalize x for numerical stability
    x_mean, x_std = np.mean(x_clean), np.std(x_clean)
    if x_std > 0:
        x_norm = (x_clean - x_mean) / x_std
    else:
        x_norm = x_clean - x_mean
    """
    # Try fitting with decreasing polynomial degrees
    for deg in range(degree, 0, -1):
        try:
            poly = np.polyfit(x_clean, y_clean, deg)
            return poly
        except np.linalg.LinAlgError:
            warnings.warn(f"SVD did not converge in Linear Least Squares")
            continue

    # Ultimate fallback: return mean
    return np.array([np.mean(y_clean)])


class LocalTests(Enum):
    CURVE = 1


def run_local_test(local_test: LocalTests):
    """Run local tests for development and debugging purposes.

    These are integration tests that download real data and generate reports.
    Use for quick verification during development.
    """

    from qis.test_data import load_etf_data
    prices = load_etf_data()
    print(prices)
    overlays = ['TLT', 'GLD']
    prices = prices[['SPY']+overlays].dropna()

    if local_test == LocalTests.CURVE:

        sd_report = SmartDiversificationReport(principal_nav=prices.iloc[:, 0], overlay_navs=prices[overlays])

        # strategies_report.plot_nav()
        sd_report.plot_smart_diversification_curve(x_var=PerfStat.BEAR_SHARPE,
                                                   y_var=PerfStat.SHARPE_RF0,
                                                   title='Total Sharpe vs Bear Sharpe')
        # strategies_report.plot_smart_diversification_curve(x_var=PerfStat.VOL, y_var=PerfStat.PA_RETURN, title='Total P.A vs Vol')

        plt.show()


if __name__ == '__main__':

    run_local_test(local_test=LocalTests.CURVE)



"""
factsheet for multi strategy report for cross sectional comparison of strategies
and generating sensitivities to parameters
see example in qis.examples.factheets.multi_strategy.py
"""
# packages
import matplotlib.pyplot as plt
import numpy as np
import pybloqs as p
import pybloqs.block.table_formatters as tf
from typing import Tuple, Optional

# qis
import qis as qis
from qis import TimePeriod, PerfParams, PerfStat, BenchmarkReturnsQuantilesRegime
from qis.portfolio.multi_portfolio_data import MultiPortfolioData
from qis.portfolio.reports.config import (PERF_PARAMS, regime_classifier, KWARGS_SUPTITLE, KWARGS_TITLE,
                                          KWARGS_TEXT, KWARGS_FOOTNOTE, KWARGS_FIG)


def generate_multi_portfolio_factsheet_with_pyblogs(multi_portfolio_data: MultiPortfolioData,
                                                    time_period: TimePeriod = None,
                                                    time_period_last: TimePeriod = None,
                                                    perf_params: PerfParams = PERF_PARAMS,
                                                    regime_classifier: BenchmarkReturnsQuantilesRegime = regime_classifier,
                                                    benchmark: str = None,
                                                    backtest_name: str = None,
                                                    heatmap_freq: str = 'YE',
                                                    param_name: Optional[str] = 'gamma',
                                                    fontsize: int = 4,
                                                    **kwargs
                                                    ) -> p.VStack:
    """
    this one is well suited for large backtests with over 100 strategies
    """
    if time_period is None:
        raise ValueError(f"pass non none time_period")
    if time_period_last is None:
        time_period_last = qis.get_time_period_shifted_by_years(time_period=time_period, n_years=1)


    if benchmark is None:
        benchmark = multi_portfolio_data.benchmark_prices.columns[0]

    plot_kwargs = dict(fontsize=fontsize,
                       linewidth=0.5,
                       digits_to_show=1, sharpe_digits=2,
                       weight='normal',
                       markersize=1,
                       framealpha=0.75)
    kwargs = qis.update_kwargs(kwargs, plot_kwargs)

    # start building blocks
    blocks = [p.Paragraph(f"{backtest_name}", **KWARGS_SUPTITLE)]

    # 1. ra performance table
    column_header = 'Strategy'
    ra_perf_table = multi_portfolio_data.get_ra_perf_table(benchmark=benchmark,
                                                           drop_benchmark=True,
                                                           is_convert_to_str=False,
                                                           perf_params=perf_params,
                                                           time_period=time_period,
                                                           column_header=column_header,
                                                           **kwargs)
    ra_perf_table_last = multi_portfolio_data.get_ra_perf_table(benchmark=benchmark,
                                                                drop_benchmark=True,
                                                                is_convert_to_str=False,
                                                                perf_params=perf_params,
                                                                time_period=time_period_last,
                                                                column_header=column_header,
                                                                **kwargs)

    # highlight
    fmt_highlight_base = tf.FmtHighlightText(rows=[ra_perf_table.index[0]], bold=False, apply_to_header_and_index=False)
    fmt_highlight_index = tf.FmtHighlightText(columns=[column_header], bold=False, italic=False,
                                              font_color=tf.colors.DARK_BLUE, apply_to_header_and_index=True)

    tables = {time_period.to_str(): ra_perf_table, time_period_last.to_str(): ra_perf_table_last}

    b_ra_perf_tables = []
    for key, table in tables.items():
        b_ra_perf_table = p.Block([p.Paragraph(f"Risk-adjusted Performance table for {key}", **KWARGS_TITLE),
                                    p.Block(table,
                                            formatters=[
                                                tf.FmtPercent(n_decimals=2, columns=[PerfStat.TOTAL_RETURN.to_str(),
                                                                                     PerfStat.PA_RETURN.to_str(),
                                                                                     PerfStat.VOL.to_str(),
                                                                                     PerfStat.MAX_DD.to_str(),
                                                                                     PerfStat.ALPHA_AN.to_str(),
                                                                                     PerfStat.R2.to_str()
                                                                                     ], apply_to_header_and_index=False),
                                                fmt_highlight_base,
                                                fmt_highlight_index,
                                                tf.FmtReplaceNaN(value=''),
                                                tf.FmtHeatmap(columns=[PerfStat.PA_RETURN.to_str(),
                                                                       PerfStat.SHARPE_EXCESS.to_str(),
                                                                       PerfStat.ALPHA_AN.to_str(),
                                                                       PerfStat.BETA.to_str()]),
                                                tf.FmtHeatmap(columns=[PerfStat.MAX_DD.to_str(),
                                                                       PerfStat.SKEWNESS.to_str()], max_color=(255,0,255)),
                                                tf.FmtAddCellBorder(each=1.0,
                                                                    columns=ra_perf_table.columns.to_list()[:1],
                                                                    color=tf.colors.GREY,
                                                                    apply_to_header_and_index=True)
                                            ]),
                                   p.Paragraph(f"  ", **KWARGS_FOOTNOTE)],
                                  **KWARGS_TEXT)
        b_ra_perf_tables.append(b_ra_perf_table)
    blocks.append([p.HStack(b_ra_perf_tables, cascade_cfg=False),
                   p.Paragraph(f" beta to {benchmark} ", **KWARGS_FOOTNOTE)])

    # 2. performance bars
    perf_columns = [PerfStat.SHARPE_EXCESS, PerfStat.MAX_DD, PerfStat.BETA]
    fig_size = qis.get_df_table_size(df=ra_perf_table)
    fig_perf_bar, axs = plt.subplots(1, len(perf_columns), figsize=(12, 1.2*fig_size[1]), constrained_layout=True)
    for idx, perf_column in enumerate(perf_columns):
        df = ra_perf_table[perf_column.to_str()].to_frame()
        colors = qis.compute_heatmap_colors(a=df.to_numpy())
        qis.plot_vbars(df=df,
                       var_format=perf_column.to_format(**kwargs),
                       title=perf_column.to_str(**kwargs),
                       legend_loc=None,
                       colors=colors,
                       is_category_names_colors=False,
                       ax=axs[idx],
                       **qis.update_kwargs(kwargs, dict(fontsize=8))
                       )
    b_fig_perf_bar = p.Block(
        [p.Paragraph(f"Performance statistics", **KWARGS_TITLE),
         p.Block(fig_perf_bar, **KWARGS_FIG)], **KWARGS_TEXT)
    blocks.append(b_fig_perf_bar)

    # 3. regime conditional
    fig_size = qis.get_df_table_size(df=ra_perf_table)
    fig_perf_regime, axs = plt.subplots(1, len(multi_portfolio_data.benchmark_prices.columns),
                                        figsize=(12, 1.2*fig_size[1]), constrained_layout=True)
    if len(multi_portfolio_data.benchmark_prices.columns) == 1:
        axs = [axs]
    for idx, benchmark in enumerate(multi_portfolio_data.benchmark_prices.columns):
        multi_portfolio_data.plot_regime_data(time_period=time_period,
                                              perf_params=perf_params,
                                              regime_classifier=regime_classifier,
                                              is_use_vbar=True,
                                              title=f"{benchmark}",
                                              benchmark=benchmark,
                                              ax=axs[idx],
                                              **qis.update_kwargs(kwargs, dict(fontsize=8)))
    b_fig_regime = p.Block(
        [p.Paragraph(f"Sharpe ratio split to {str(benchmark)} Bear/Normal/Bull {regime_classifier.freq}-freq regimes", **KWARGS_TITLE),
         p.Block(fig_perf_regime, **KWARGS_FIG)], **KWARGS_TEXT)
    blocks.append(b_fig_regime)

    # 4. heatmap of annual returns
    prices = multi_portfolio_data.get_navs(time_period=time_period)
    returns_ye = qis.compute_periodic_returns(prices=prices,
                                              freq=heatmap_freq,
                                              time_period=time_period,
                                              total_name='Total',
                                              add_total=True,
                                              **qis.update_kwargs(kwargs, dict(date_format='%Y'))
                                              ).T
    fmt_highlight_base = tf.FmtHighlightText(rows=[returns_ye.index[0]], bold=False, apply_to_header_and_index=False)
    fmt_highlight_index = tf.FmtHighlightText(columns=[returns_ye.index.name], bold=False, italic=False,
                                              font_color=tf.colors.DARK_BLUE, apply_to_header_and_index=True)

    b_returns_ye = p.Block([p.Paragraph(f"Annual returns table", **KWARGS_TITLE),
                            p.Block(returns_ye,
                                    formatters=[tf.FmtPercent(n_decimals=0, columns=returns_ye.columns, apply_to_header_and_index=False),
                                                fmt_highlight_base,
                                                fmt_highlight_index,
                                                tf.FmtReplaceNaN(value=''),
                                                tf.FmtHeatmap(columns=returns_ye.columns, axis=0)
                                                ]),
                            p.Paragraph(f"  ", **KWARGS_FOOTNOTE)],
                           **KWARGS_TEXT)
    blocks.append(b_returns_ye)

    # param sensetivity scatter
    if param_name is not None:
        ra_perf_table = multi_portfolio_data.get_ra_perf_table(benchmark=benchmark,
                                                               drop_benchmark=True,
                                                               is_convert_to_str=False,
                                                               perf_params=perf_params,
                                                               time_period=time_period,
                                                               column_header=column_header,
                                                               **kwargs)

        if 'base' in ra_perf_table.index:
            ra_perf_table = ra_perf_table.drop('base', axis=0)

        hue_name = 'group'

        xy = ra_perf_table.copy()
        xy[hue_name] = [x.split(' ')[0] for x in ra_perf_table.index]
        xy[param_name] = [float(x.split('=')[-1]) for x in ra_perf_table.index]
        x_limits = (np.nanmin(xy[param_name]), np.nanmax(xy[param_name]))

        fig_scatter1, ax = plt.subplots(1, 1, figsize=(14, 4.5), constrained_layout=True)
        qis.plot_scatter(df=xy, x=param_name, y=PerfStat.SHARPE_EXCESS.to_str(), hue=hue_name,
                         title=f"Sharpe",
                         var_format='{:.2f}',
                         x_limits=x_limits,
                         ax=ax,
                         **kwargs)
        b_fig_scatter1 = p.Block([p.Paragraph(f"Scatter of {PerfStat.SHARPE_EXCESS.to_str()} for {param_name} sensitivity", **KWARGS_TITLE),
                                p.Block(fig_scatter1, **KWARGS_FIG)],
                               **KWARGS_TEXT)
        blocks.append(b_fig_scatter1)

        fig_scatter2, ax = plt.subplots(1, 1, figsize=(14, 4.5), constrained_layout=True)
        qis.plot_scatter(df=xy, x=param_name, y=PerfStat.MAX_DD.to_str(), hue=hue_name,
                         title=f"Max DD",
                         var_format='{:.2%}',
                         x_limits=x_limits,
                         ax=ax,
                         **kwargs)
        b_fig_scatter = p.Block([p.Paragraph(f"Scatter of {PerfStat.MAX_DD.to_str(**kwargs)} for {param_name} sensitivity", **KWARGS_TITLE),
                                 p.Block(fig_scatter2, **KWARGS_FIG)],
                                **KWARGS_TEXT)
        blocks.append(b_fig_scatter)

    report = p.VStack(blocks, cascade_cfg=False)
    return report


def generate_multi_portfolio_factsheet(multi_portfolio_data: MultiPortfolioData,
                                       time_period: TimePeriod = None,
                                       perf_params: PerfParams = PERF_PARAMS,
                                       regime_classifier: BenchmarkReturnsQuantilesRegime = regime_classifier,
                                       regime_benchmark: str = None,
                                       backtest_name: str = None,
                                       heatmap_freq: str = 'YE',
                                       figsize: Tuple[float, float] = (8.3, 11.7),  # A4 for portrait
                                       is_grouped: bool = False,
                                       fontsize: int = 5,
                                       **kwargs
                                       ) -> plt.Figure:
    """
    for portfolio data with structurally different strategies
    for portfolios with large universe use is_grouped = True to report tunrover and exposures by groups
    """
    if regime_benchmark is None:
        regime_benchmark = multi_portfolio_data.benchmark_prices.columns[0]

    plot_kwargs = dict(fontsize=fontsize,
                       linewidth=0.5,
                       digits_to_show=1, sharpe_digits=2,
                       weight='normal',
                       markersize=1,
                       framealpha=0.75)
    kwargs = qis.update_kwargs(kwargs, plot_kwargs)

    fig = plt.figure(figsize=figsize, constrained_layout=True)
    gs = fig.add_gridspec(nrows=7, ncols=4, wspace=0.0, hspace=0.0)

    if backtest_name is not None:
        fig.suptitle(backtest_name, fontweight="bold", fontsize=8, color='blue')

    multi_portfolio_data.plot_nav(ax=fig.add_subplot(gs[0, :2]),
                                  time_period=time_period,
                                  regime_benchmark=regime_benchmark,
                                  perf_params=perf_params,
                                  regime_classifier=regime_classifier,
                                  title='Cumulative performance',
                                  **kwargs)

    multi_portfolio_data.plot_drawdowns(ax=fig.add_subplot(gs[1, :2]),
                                        time_period=time_period,
                                        regime_benchmark=regime_benchmark,
                                        dd_legend_type=qis.DdLegendType.DETAILED,
                                        regime_classifier=regime_classifier,
                                        title='Running Drawdowns',
                                        **kwargs)

    multi_portfolio_data.plot_rolling_time_under_water(ax=fig.add_subplot(gs[2, :2]),
                                                       time_period=time_period,
                                                       regime_benchmark=regime_benchmark,
                                                       regime_classifier=regime_classifier,
                                                       title='Rolling time under water',
                                                       **kwargs)

    multi_portfolio_data.plot_exposures(ax=fig.add_subplot(gs[3, :2]),
                                        portfolio_idx=0,
                                        time_period=time_period,
                                        regime_benchmark=regime_benchmark,
                                        regime_classifier=regime_classifier,
                                        **kwargs)

    multi_portfolio_data.plot_turnover(ax=fig.add_subplot(gs[4, :2]),
                                       time_period=time_period,
                                       regime_benchmark=regime_benchmark,
                                       regime_classifier=regime_classifier,
                                       **kwargs)

    multi_portfolio_data.plot_costs(ax=fig.add_subplot(gs[5, :2]),
                                    time_period=time_period,
                                    regime_benchmark=regime_benchmark,
                                    regime_classifier=regime_classifier,
                                    **kwargs)

    # select two benchmarks for factor exposures
    multi_portfolio_data.plot_factor_betas(axs=[fig.add_subplot(gs[6, :2]), fig.add_subplot(gs[6, 2:])],
                                           benchmark_prices=multi_portfolio_data.benchmark_prices.iloc[:, :2],
                                           time_period=time_period,
                                           regime_benchmark=regime_benchmark,
                                           regime_classifier=regime_classifier,
                                           **kwargs)

    multi_portfolio_data.plot_performance_bars(ax=fig.add_subplot(gs[0, 2]),
                                               perf_params=perf_params,
                                               perf_column=PerfStat.SHARPE_EXCESS,
                                               time_period=time_period,
                                               **qis.update_kwargs(kwargs, dict(fontsize=fontsize)))

    multi_portfolio_data.plot_performance_bars(ax=fig.add_subplot(gs[0, 3]),
                                               perf_params=perf_params,
                                               perf_column=PerfStat.MAX_DD,
                                               time_period=time_period,
                                               **qis.update_kwargs(kwargs, dict(fontsize=fontsize)))

    multi_portfolio_data.plot_ra_perf_table(ax=fig.add_subplot(gs[1, 2:]),
                                            perf_params=perf_params,
                                            time_period=time_period,
                                            **qis.update_kwargs(kwargs, dict(fontsize=fontsize)))

    multi_portfolio_data.plot_periodic_returns(ax=fig.add_subplot(gs[2, 2:]),
                                               heatmap_freq=heatmap_freq,
                                               title=f"{heatmap_freq} returns",
                                               time_period=time_period,
                                               **qis.update_kwargs(kwargs, dict(fontsize=fontsize)))

    """
    multi_portfolio_data.plot_ra_perf_table(ax=fig.add_subplot(gs[3, 2:]),
                                            perf_params=perf_params,
                                            time_period=qis.get_time_period_shifted_by_years(time_period=time_period),
                                            **qis.update_kwargs(kwargs, dict(fontsize=fontsize, freq_reg='W-WED')))
    """
    multi_portfolio_data.plot_corr_table(ax=fig.add_subplot(gs[3, 2:]),
                                         time_period=time_period,
                                         freq='W-WED',
                                         **qis.update_kwargs(kwargs, dict(fontsize=fontsize)))

    multi_portfolio_data.plot_regime_data(ax=fig.add_subplot(gs[4, 2:]),
                                          is_grouped=False,
                                          time_period=time_period,
                                          perf_params=perf_params,
                                          regime_classifier=regime_classifier,
                                          benchmark=multi_portfolio_data.benchmark_prices.columns[0],
                                          **kwargs)
    multi_portfolio_data.plot_regime_data(ax=fig.add_subplot(gs[5, 2:]),
                                          is_grouped=False,
                                          time_period=time_period,
                                          perf_params=perf_params,
                                          regime_classifier=regime_classifier,
                                          benchmark=multi_portfolio_data.benchmark_prices.columns[1],
                                          **kwargs)
    """
    multi_portfolio_data.plot_returns_scatter(ax=fig.add_subplot(gs[5, 2:]),
                                              time_period=time_period,
                                              benchmark=regime_benchmark,
                                              **kwargs)
    """

    return fig



"""
generate strategy factsheet report with comparision to benchmark strategy using MultiPortfolioData object
for test implementation see qis.examples.portfolio_factsheet
"""
import numpy as np
# packages
import pandas as pd
import seaborn as sns
import matplotlib.pyplot as plt
from typing import List, Tuple, Optional
import qis as qis
from qis import TimePeriod, PerfParams, BenchmarkReturnsQuantilesRegime

# portfolio
from qis.portfolio.portfolio_data import AttributionMetric
from qis.portfolio.multi_portfolio_data import MultiPortfolioData
from qis.portfolio.reports.strategy_factsheet import generate_strategy_factsheet
from qis.portfolio.reports.config import PERF_PARAMS, regime_classifier


def generate_strategy_benchmark_factsheet_plt(multi_portfolio_data: MultiPortfolioData,
                                              strategy_idx: int = 0,  # strategy is multi_portfolio_data[strategy_idx]
                                              benchmark_idx: int = 1,  # benchmark is multi_portfolio_data[benchmark_idx]
                                              time_period: TimePeriod = None,
                                              perf_params: PerfParams = PERF_PARAMS,
                                              regime_classifier: BenchmarkReturnsQuantilesRegime = BenchmarkReturnsQuantilesRegime(),
                                              backtest_name: str = None,
                                              add_benchmarks_to_navs: bool = False,
                                              add_brinson_attribution: bool = True,
                                              add_exposures_pnl_attribution: bool = False,
                                              add_strategy_factsheet: bool = False,
                                              add_grouped_exposures: bool = False,  # for strategy factsheet
                                              add_grouped_cum_pnl: bool = False,  # for strategy factsheet
                                              add_tracking_error_table: bool = False,
                                              add_exposures_comp: bool = False,
                                              is_grouped: Optional[bool] = True,
                                              figsize: Tuple[float, float] = (8.3, 11.7),  # A4 for portrait
                                              fontsize: int = 5,
                                              heatmap_fontsize: int = 4,
                                              add_joint_instrument_history_report: bool = False,
                                              **kwargs
                                              ) -> List[plt.Figure]:
    """
    report designed for 1 strategy and 1 benchmark report using matplotlib figure
    multi_portfolio_data = [stragegy portfolio, benchmark strategy portfolio]
    is_grouped sets report of annual returns by asset classes or portfolio instruments if their number is less than 10
    1 page: generate the stragegy portfolio factsheet
    2 page: generate the comparision to benchmark
    """
    if len(multi_portfolio_data.portfolio_datas) == 1:
        raise ValueError(f"must be at least two strategies")

    if is_grouped is None:
        if len(multi_portfolio_data.portfolio_datas[0].get_weights().columns) >= 10:  # otherwise tables look too bad
            is_grouped = True
        else:
            is_grouped = False
    else:
        if is_grouped is False:
            if len(multi_portfolio_data.portfolio_datas[0].get_weights().columns) >= 10:  # tables look too bad
                is_grouped = True

    # set reporting time period here
    if time_period is None:
        time_period = qis.get_time_period(multi_portfolio_data.portfolio_datas[0].get_portfolio_nav())

    # set report specific kqargs
    plot_kwargs = dict(fontsize=fontsize,
                       linewidth=0.5,
                       digits_to_show=1, sharpe_digits=2,
                       weight='normal',
                       markersize=1,
                       framealpha=0.75,
                       time_period=time_period)
    kwargs = qis.update_kwargs(kwargs, plot_kwargs)

    figs = []
    fig = plt.figure(figsize=figsize, constrained_layout=True)
    figs.append(fig)
    gs = fig.add_gridspec(nrows=14, ncols=4, wspace=0.0, hspace=0.0)

    if backtest_name is not None:
        fig.suptitle(backtest_name, fontweight="bold", fontsize=8, color='blue')

    regime_benchmark = multi_portfolio_data.benchmark_prices.columns[0]
    benchmark_price = multi_portfolio_data.benchmark_prices[regime_benchmark]

    multi_portfolio_data.plot_nav(ax=fig.add_subplot(gs[:2, :2]),
                                  add_benchmarks_to_navs=add_benchmarks_to_navs,
                                  regime_benchmark=regime_benchmark,
                                  perf_params=perf_params,
                                  regime_classifier=regime_classifier,
                                  title=f"Cumulative performance with background colors using bear/normal/bull regimes of {regime_benchmark} {regime_classifier.freq}-returns",
                                  **kwargs)

    multi_portfolio_data.plot_drawdowns(ax=fig.add_subplot(gs[2:4, :2]),
                                        add_benchmarks_to_navs=add_benchmarks_to_navs,
                                        regime_benchmark=regime_benchmark,
                                        regime_classifier=regime_classifier,
                                        title='Running Drawdowns',
                                        **kwargs)

    multi_portfolio_data.plot_rolling_time_under_water(ax=fig.add_subplot(gs[4:6, :2]),
                                                       add_benchmarks_to_navs=add_benchmarks_to_navs,
                                                       regime_benchmark=regime_benchmark,
                                                       regime_classifier=regime_classifier,
                                                       title='Rolling time under water',
                                                       **kwargs)

    multi_portfolio_data.plot_rolling_perf(ax=fig.add_subplot(gs[6:8, :2]),
                                           add_benchmarks_to_navs=add_benchmarks_to_navs,
                                           regime_benchmark=regime_benchmark,
                                           regime_classifier=regime_classifier,
                                           **kwargs)

    multi_portfolio_data.plot_exposures(ax=fig.add_subplot(gs[8:10, :2]),
                                        regime_benchmark=regime_benchmark,
                                        regime_classifier=regime_classifier,
                                        **kwargs)

    multi_portfolio_data.plot_turnover(ax=fig.add_subplot(gs[10:12, :2]),
                                       regime_benchmark=regime_benchmark,
                                       regime_classifier=regime_classifier,
                                       **kwargs)

    multi_portfolio_data.plot_costs(ax=fig.add_subplot(gs[12:14, :2]),
                                    regime_benchmark=regime_benchmark,
                                    regime_classifier=regime_classifier,
                                    **kwargs)

    multi_portfolio_data.plot_ac_ra_perf_table(ax=fig.add_subplot(gs[0:2, 2:]),
                                               benchmark_price=benchmark_price,
                                               add_benchmarks_to_navs=add_benchmarks_to_navs,
                                               perf_params=perf_params,
                                               is_grouped=is_grouped,
                                               **qis.update_kwargs(kwargs, dict(fontsize=fontsize)))

    time_period1 = qis.get_time_period_shifted_by_years(time_period=time_period, n_years=1)
    # change regression to weekly
    if pd.infer_freq(benchmark_price.index) in ['B', 'D']:
        local_kwargs = qis.update_kwargs(kwargs, dict(time_period=time_period1, freq_reg='W-WED', fontsize=fontsize))
    else:
        local_kwargs = qis.update_kwargs(kwargs, dict(time_period=time_period1, fontsize=fontsize))

    multi_portfolio_data.plot_ac_ra_perf_table(ax=fig.add_subplot(gs[2:4, 2:]),
                                               benchmark_price=benchmark_price,
                                               add_benchmarks_to_navs=add_benchmarks_to_navs,
                                               perf_params=perf_params,
                                               is_grouped=is_grouped,
                                               **local_kwargs)

    # periodic returns
    local_kwargs = qis.update_kwargs(kwargs=kwargs,
                                     new_kwargs=dict(fontsize=heatmap_fontsize, square=False,
                                                     x_rotation=90, transpose=True))

    # periodic returns
    multi_portfolio_data.plot_periodic_returns(add_benchmarks_to_navs=add_benchmarks_to_navs,
                                                                  ax=fig.add_subplot(gs[4:6, 2:4]),
                                                                  **local_kwargs)

    # regime data for 1 or two benchmarks
    if len(multi_portfolio_data.benchmark_prices.columns) == 1:
        benchmark_prices_ = {benchmark_price.name: fig.add_subplot(gs[6:8, 2:4])}
    else:
        benchmark_prices_ = {multi_portfolio_data.benchmark_prices.columns[0]: fig.add_subplot(gs[6:8, 2]),
                             multi_portfolio_data.benchmark_prices.columns[1]: fig.add_subplot(gs[6:8, 3])}
    for benchmark_, ax_ in benchmark_prices_.items():
        post_title = f"Sharpe ratio in {benchmark_} Bear/Normal/Bull {regime_classifier.freq}-freq regimes"
        multi_portfolio_data.plot_regime_data(benchmark=benchmark_,
                                              add_benchmarks_to_navs=add_benchmarks_to_navs,
                                              is_grouped=False,
                                              title=f"{post_title}",
                                              perf_params=perf_params,
                                              regime_classifier=regime_classifier,
                                              ax=ax_,
                                              **qis.update_kwargs(kwargs, dict(fontsize=fontsize, x_rotation=90)))

    # vol regimes
    """
    multi_portfolio_data.portfolio_datas[0].plot_vol_regimes(ax=fig.add_subplot(gs[8:10, 2]),
                                                             benchmark_price=benchmark_price,
                                                             perf_params=perf_params,
                                                             freq=regime_classifier.freq,
                                                             **qis.update_kwargs(kwargs, dict(fontsize=fontsize, x_rotation=90)))
    multi_portfolio_data.portfolio_datas[1].plot_vol_regimes(ax=fig.add_subplot(gs[8:10, 3]),
                                                             benchmark_price=benchmark_price,
                                                             freq=perf_params.freq,
                                                             regime_classifier=regime_classifier,
                                                             **qis.update_kwargs(kwargs, dict(fontsize=fontsize, x_rotation=90)))
    """
    multi_portfolio_data.plot_instrument_pnl_diff(ax=fig.add_subplot(gs[8:10, 2:]),
                                                  regime_benchmark=regime_benchmark,
                                                  regime_classifier=regime_classifier,
                                                  **kwargs)

    # plot beta to the regime_benchmark
    multi_portfolio_data.plot_factor_betas(axs=[fig.add_subplot(gs[10:12, 2:])],
                                           benchmark_prices=multi_portfolio_data.benchmark_prices[regime_benchmark].to_frame(),
                                           regime_benchmark=regime_benchmark,
                                           regime_classifier=regime_classifier,
                                           **kwargs)
    with sns.axes_style('whitegrid'):
        multi_portfolio_data.plot_returns_scatter(ax=fig.add_subplot(gs[12:, 2:]),
                                                  benchmark=regime_benchmark,
                                                  **qis.update_kwargs(kwargs, dict(freq=perf_params.freq_reg)))

    if add_brinson_attribution:
        with sns.axes_style("darkgrid"):
            fig1 = plt.figure(figsize=figsize, constrained_layout=True)
            this_name = backtest_name if backtest_name is not None else ''
            fig1.suptitle(f'{this_name} Brinson performance attribution report', fontweight="bold", fontsize=8, color='blue')
            figs.append(fig1)
            gs = fig1.add_gridspec(nrows=3, ncols=2, wspace=0.0, hspace=0.0)
            axs = [fig1.add_subplot(gs[0, 0]), fig1.add_subplot(gs[0, 1]),
                   fig1.add_subplot(gs[1, 0]), fig1.add_subplot(gs[1, 1]),
                   fig1.add_subplot(gs[2, 1])]
            multi_portfolio_data.plot_brinson_attribution(strategy_idx=strategy_idx,
                                                          benchmark_idx=benchmark_idx,
                                                          freq=None,
                                                          axs=axs,
                                                          total_column='Total Sum',
                                                          is_exclude_interaction_term=True,
                                                          **kwargs)
        # add exposure diff
        multi_portfolio_data.plot_exposures_diff(ax=fig1.add_subplot(gs[2, 0]),
                                                 regime_benchmark=regime_benchmark,
                                                 regime_classifier=regime_classifier,
                                                 **kwargs)

    if add_tracking_error_table:
        tre_table = multi_portfolio_data.compute_tracking_error_table(strategy_idx=strategy_idx,
                                                                      benchmark_idx=benchmark_idx,
                                                                      **kwargs)
        fig1, ax = plt.subplots(1, 1, figsize=qis.get_df_table_size(df=tre_table), constrained_layout=True)
        fig1.suptitle(f'{backtest_name} Tracking error table', fontweight="bold", fontsize=8, color='blue')
        figs.append(fig1)
        qis.plot_df_table(df=tre_table,
                          first_row_height=0.075,
                          var_format='{:.2%}',
                          rotation_for_columns_headers=90,
                          heatmap_columns=[x+1 for x in np.arange(len(tre_table.columns))],
                          ax=ax,
                          cmap='Blues',
                          **kwargs)

    if add_exposures_comp:
        strategy_name = multi_portfolio_data.portfolio_datas[strategy_idx].ticker
        benchmark_name = multi_portfolio_data.portfolio_datas[benchmark_idx].ticker
        strategy_weights, benchmark_weights = multi_portfolio_data.get_aligned_weights(strategy_idx=strategy_idx,
                                                                                       benchmark_idx=benchmark_idx,
                                                                                       is_grouped=is_grouped,
                                                                                       **kwargs)

        strategy_turnover, benchmark_turnover = multi_portfolio_data.get_aligned_turnover(strategy_idx=strategy_idx,
                                                                                          benchmark_idx=benchmark_idx,
                                                                                          is_grouped=is_grouped,
                                                                                          **kwargs)
        for inst in strategy_weights.columns:
            df1 = pd.concat([strategy_weights[inst].rename(strategy_name),
                            benchmark_weights[inst].rename(benchmark_name)], axis=1)
            df2 = pd.concat([strategy_turnover[inst].rename(strategy_name),
                            benchmark_turnover[inst].rename(benchmark_name)], axis=1)
            with sns.axes_style("darkgrid"):
                fig1, axs = plt.subplots(2, 1, figsize=figsize, constrained_layout=True)
                fig1.suptitle(f'{inst} Exposures and Turnover', fontweight="bold", fontsize=8, color='blue')
                figs.append(fig1)
                qis.plot_time_series(df=df1, title='Exposures',
                                     legend_stats=qis.LegendStats.AVG_MIN_MAX_LAST,
                                     var_format='{:,.2%}',
                                     ax=axs[0], **kwargs)
                qis.plot_time_series(df=df2, title='Turnover',
                                     legend_stats=qis.LegendStats.AVG_MIN_MAX_LAST,
                                     var_format='{:,.2%}',
                                     ax=axs[1], **kwargs)

            plt.close('all')

    if add_exposures_pnl_attribution:
        strategy_name = multi_portfolio_data.portfolio_datas[strategy_idx].ticker
        benchmark_name = multi_portfolio_data.portfolio_datas[benchmark_idx].ticker
        strategy_grouped_exposures_agg, strategy_grouped_exposures_by_inst = \
            multi_portfolio_data.portfolio_datas[strategy_idx].get_grouped_long_short_weights(time_period=time_period)

        benchmark_grouped_exposures_agg, benchmark_grouped_exposures_by_inst = \
            multi_portfolio_data.portfolio_datas[benchmark_idx].get_grouped_long_short_weights(time_period=time_period)

        strategy_grouped_pnls_agg, strategy_grouped_pnls_by_inst \
            = multi_portfolio_data.portfolio_datas[strategy_idx].get_grouped_cum_pnls(time_period=time_period)

        benchmark_grouped_pnls_agg, strategy_grouped_pnls_by_inst \
            = multi_portfolio_data.portfolio_datas[benchmark_idx].get_grouped_cum_pnls(time_period=time_period)

        nrows = len(strategy_grouped_exposures_agg.keys())
        fig2 = plt.figure(figsize=figsize, constrained_layout=True)
        figs.append(fig2)
        fig2.suptitle(f"{backtest_name}: Strategy vs Benchmark exposures by groups for period {time_period.to_str()}",
                     fontweight="bold", fontsize=8, color='blue')
        gs = fig2.add_gridspec(nrows=nrows, ncols=2, wspace=0.0, hspace=0.0)
        local_kwargs = qis.update_kwargs(kwargs=kwargs, new_kwargs=dict(framealpha=0.9))
        for idx, (group, exposures_agg) in enumerate(strategy_grouped_exposures_agg.items()):
            df1 = pd.concat([strategy_grouped_exposures_agg[group]['Total'].rename(strategy_name),
                             benchmark_grouped_exposures_agg[group]['Total'].rename(benchmark_name)],
                            axis=1)
            df2 = pd.concat([strategy_grouped_pnls_agg[group]['Total'].rename(strategy_name),
                             benchmark_grouped_pnls_agg[group]['Total'].rename(benchmark_name)],
                            axis=1)

            datas = {f"{group} aggregated net exposures": df1,
                     f"{group} cumulative P&L attribution": df2}
            for idx_, (key, df) in enumerate(datas.items()):
                ax = fig2.add_subplot(gs[idx, idx_])
                qis.plot_time_series(df=df,
                                     var_format='{:,.0%}',
                                     legend_stats=qis.LegendStats.AVG_MIN_MAX_LAST,
                                     title=f"{key}",
                                     ax=ax,
                                     **local_kwargs)
                multi_portfolio_data.add_regime_shadows(ax=ax, regime_benchmark=regime_benchmark, index=df.index, regime_classifier=regime_classifier)
                qis.set_spines(ax=ax, bottom_spine=False, left_spine=False)

    if add_joint_instrument_history_report:
        perf_columns = (qis.PerfStat.START_DATE, qis.PerfStat.END_DATE, qis.PerfStat.PA_RETURN,
                        qis.PerfStat.VOL, qis.PerfStat.SHARPE_RF0,
                        qis.PerfStat.MAX_DD, qis.PerfStat.MAX_DD_VOL, qis.PerfStat.SKEWNESS)

        joint_prices = pd.concat([multi_portfolio_data.portfolio_datas[strategy_idx].prices,
                                  multi_portfolio_data.portfolio_datas[benchmark_idx].prices],
                                 axis=1)
        joint_prices = joint_prices.loc[:, ~joint_prices.columns.duplicated(keep='first')]
        fig = qis.generate_price_history_report(prices=joint_prices,
                                                **qis.update_kwargs(kwargs, dict(fontsize=4, figsize=figsize,
                                                                                 perf_columns=perf_columns)))
        fig.suptitle('Instruments in Joint Universe', fontweight="bold", fontsize=8, color='blue')
        figs.append(fig)

    if add_strategy_factsheet:
        for portfolio_data in multi_portfolio_data.portfolio_datas:
            figs.append(generate_strategy_factsheet(portfolio_data=portfolio_data,
                                                    benchmark_prices=multi_portfolio_data.benchmark_prices,
                                                    perf_params=perf_params,
                                                    regime_classifier=regime_classifier,
                                                    add_grouped_exposures=add_grouped_exposures,
                                                    add_grouped_cum_pnl=add_grouped_cum_pnl,
                                                    is_grouped=is_grouped,
                                                    **kwargs  # time period will be in kwargs
                                                    ))
        figs = qis.to_flat_list(figs)

    return figs


def generate_strategy_benchmark_active_perf_plt(multi_portfolio_data: MultiPortfolioData,
                                                strategy_idx: int = 0,  # strategy is multi_portfolio_data[strategy_idx]
                                                benchmark_idx: int = 1, # benchmark is multi_portfolio_data[benchmark_idx]
                                                time_period: TimePeriod = None,
                                                perf_params: PerfParams = PERF_PARAMS,
                                                regime_classifier: BenchmarkReturnsQuantilesRegime = BenchmarkReturnsQuantilesRegime(),
                                                backtest_name: str = None,
                                                weight_freq: Optional[str] = 'ME',
                                                add_strategy_factsheet: bool = False,
                                                add_grouped_exposures: bool = False,  # for strategy factsheet
                                                add_grouped_cum_pnl: bool = False,  # for strategy factsheet
                                                figsize: Tuple[float, float] = (8.3, 11.7),  # A4 for portrait
                                                is_long_only: bool = False,
                                                fontsize: int = 8,
                                                **kwargs
                                                ) -> List[plt.Figure]:
    """
    display 2*2 plot with nav and Brinson attribution
    """
    if len(multi_portfolio_data.portfolio_datas) == 1:
        raise ValueError(f"must be at least two strategies")

    # set report specific kqargs
    plot_kwargs = dict(fontsize=fontsize,
                       linewidth=1.0,
                       digits_to_show=1, sharpe_digits=2,
                       weight='normal',
                       markersize=1,
                       framealpha=0.75,
                       time_period=time_period,
                       perf_stats_labels=qis.PerfStatsLabels.DETAILED_WITH_DD.value)
    kwargs = qis.update_kwargs(kwargs, plot_kwargs)

    figs = []
    fig = plt.figure(figsize=figsize, constrained_layout=True)
    figs.append(fig)
    gs = fig.add_gridspec(nrows=3, ncols=2, wspace=0.0, hspace=0.0)

    if backtest_name is not None:
        fig.suptitle(backtest_name, fontweight="bold", fontsize=8, color='blue')

    regime_benchmark = multi_portfolio_data.benchmark_prices.columns[0]
    # benchmark_price = multi_portfolio_data.benchmark_prices[regime_benchmark]

    multi_portfolio_data.plot_nav(ax=fig.add_subplot(gs[0, 0]),
                                  regime_benchmark=regime_benchmark,
                                  perf_params=perf_params,
                                  regime_classifier=regime_classifier,
                                  title='Cumulative performance',
                                  **kwargs)

    totals_table, active_total, grouped_allocation_return, grouped_selection_return, grouped_interaction_return = \
        multi_portfolio_data.compute_brinson_attribution(strategy_idx=strategy_idx,
                                                         benchmark_idx=benchmark_idx,
                                                         freq=None,
                                                         time_period=time_period,
                                                         is_exclude_interaction_term=True)

    datas = {'Active total return (Brinson attribution)': (active_total.cumsum(0), fig.add_subplot(gs[0, 1])),
             'Asset class allocation return': (grouped_allocation_return.cumsum(0), fig.add_subplot(gs[1, 1])),
             'Instrument selection return': (grouped_selection_return.cumsum(0), fig.add_subplot(gs[2, 1]))}
    for key, (df, ax) in datas.items():
        legend_labels = [column + ', sum=' + '{:.1%}'.format(df[column].iloc[-1]) for column in df.columns]
        qis.plot_time_series(df=df,
                             title=key,
                             var_format='{:.1%}',
                             legend_labels=legend_labels,
                             ax=ax,
                             **kwargs)

        if regime_benchmark is not None:
            multi_portfolio_data.add_regime_shadows(ax=ax, regime_benchmark=regime_benchmark, index=df.index,
                                                    regime_classifier=regime_classifier)

    # weights box plot
    # make horizontal is too many instruments
    if len(multi_portfolio_data.portfolio_datas[0].prices.columns) > 9:
        local_kwargs = qis.update_kwargs(kwargs, dict(x_rotation=90, add_hue_to_legend_title=False))
    else:
        local_kwargs = qis.update_kwargs(kwargs, dict(add_hue_to_legend_title=False))
    ax = fig.add_subplot(gs[1, 0])
    multi_portfolio_data.plot_weights_boxplot(strategy_idx=strategy_idx,
                                              benchmark_idx=benchmark_idx,
                                              freq=weight_freq,
                                              is_grouped=True,
                                              title=f"Boxplot of weights by groups at {weight_freq}-freq",
                                              ax=ax,
                                              **local_kwargs)
    if is_long_only:
        qis.set_y_limits(ax=ax, y_limits=(0, None))
    ax = fig.add_subplot(gs[2, 0])
    multi_portfolio_data.plot_weights_boxplot(strategy_idx=strategy_idx,
                                              benchmark_idx=benchmark_idx,
                                              freq=weight_freq,
                                              is_grouped=False,
                                              title=f"Boxplot of weights by instruments at {weight_freq}-freq",
                                              ax=ax,
                                              **local_kwargs)
    if is_long_only:
        qis.set_y_limits(ax=ax, y_limits=(0, None))

    if add_strategy_factsheet:
        for portfolio_data in multi_portfolio_data.portfolio_datas:
            figs.append(generate_strategy_factsheet(portfolio_data=portfolio_data,
                                                    benchmark_prices=multi_portfolio_data.benchmark_prices,
                                                    perf_params=perf_params,
                                                    regime_classifier=regime_classifier,
                                                    add_grouped_exposures=add_grouped_exposures,
                                                    add_grouped_cum_pnl=add_grouped_cum_pnl,
                                                    **kwargs  # time period will be in kwargs
                                                    ))
        figs = qis.to_flat_list(figs)
    return figs


def generate_performance_attribution_report(multi_portfolio_data: MultiPortfolioData,
                                            time_period: TimePeriod = None,
                                            figsize: Tuple[float, float] = (8.3, 11.7),  # A4 for portrait
                                            **kwargs
                                            ) -> List[plt.Figure]:
    figs = []
    with sns.axes_style('darkgrid'):
        fig = plt.figure(figsize=figsize, constrained_layout=True)
        figs.append(fig)
        gs = fig.add_gridspec(nrows=3, ncols=2, wspace=0.0, hspace=0.0)

        multi_portfolio_data.plot_nav(ax=fig.add_subplot(gs[0, :]),
                                      time_period=time_period,
                                      **kwargs)

        multi_portfolio_data.plot_performance_attribution(portfolio_ids=[0],
                                                          time_period=time_period,
                                                          attribution_metric=AttributionMetric.PNL,
                                                          ax=fig.add_subplot(gs[1, :]),
                                                          **kwargs)

        multi_portfolio_data.plot_performance_attribution(portfolio_ids=[1],
                                                          time_period=time_period,
                                                          attribution_metric=AttributionMetric.PNL,
                                                          ax=fig.add_subplot(gs[2, :]),
                                                          **kwargs)

        fig, ax = plt.subplots(1, 1, figsize=figsize)
        figs.append(fig)
        multi_portfolio_data.plot_performance_periodic_table(portfolio_ids=[0],
                                                             time_period=time_period,
                                                             attribution_metric=AttributionMetric.INST_PNL,
                                                             freq='ME',
                                                             ax=ax,
                                                             **kwargs)

    return figs



"""
generate and load test data using yf library
"""

import pandas as pd
import yfinance as yf
from enum import Enum

import qis.file_utils as fu
import qis.local_path as local_path


RESOURCE_PATH = local_path.get_paths()['RESOURCE_PATH']


def load_etf_data() -> pd.DataFrame:
    prices = fu.load_df_from_csv(file_name='etf_prices', local_path=RESOURCE_PATH)
    return prices


class LocalTests(Enum):
    ETF_PRICES = 1
    TEST_LOADING = 2


def run_local_test(local_test: LocalTests):
    """Run local tests for development and debugging purposes.

    These are integration tests that download real data and generate reports.
    Use for quick verification during development.
    """

    if local_test == LocalTests.ETF_PRICES:
        prices = yf.download(tickers=['SPY', 'QQQ', 'EEM', 'TLT', 'IEF', 'LQD', 'HYG', 'SHY', 'GLD'],
                             start="2003-12-31", end=None, ignore_tz=True, auto_adjust=True)['Close']
        print(prices)
        fu.save_df_to_csv(df=prices, file_name='etf_prices', local_path=RESOURCE_PATH)

    elif local_test == LocalTests.TEST_LOADING:
        prices = load_etf_data()
        print(prices)


if __name__ == '__main__':

    run_local_test(local_test=LocalTests.ETF_PRICES)



import numpy as np
import pandas as pd
import seaborn as sns
import matplotlib.pyplot as plt
from enum import Enum
from typing import Union
import yfinance as yf
import qis


def fetch_hourly_data(ticker: str = 'BTC-USD') -> pd.DataFrame:
    """
    yf reports timestamps of bars at the start of the period: we shift it to the end of the period
    """
    asset = yf.Ticker(ticker)
    ohlc_data = asset.history(period="1y", interval="1h")
    ohlc_data.index = [t + pd.Timedelta(minutes=60) for t in ohlc_data.index]
    ohlc_data = ohlc_data.rename({'Open': 'open', 'High': 'high', 'Low': 'low', 'Close': 'close'}, axis=1)
    ohlc_data.index = pd.to_datetime(ohlc_data.index).tz_convert('UTC')
    ohlc_data.index.name = 'timestamp'
    return ohlc_data


def compute_vols(prices: pd.Series,
                 span: int = 5*24
                 ) -> pd.DataFrame:
    """
    compute rolling ewma with and without weekends
    need to adjust annualisation factor
    """
    returns = np.log(prices).diff(1)
    init_value = np.nanvar(returns, axis=0)  # set initial value to average variance
    vol = qis.compute_ewm_vol(data=returns, span=span, annualization_factor=365*24, init_value=init_value)

    returns1 = returns.where(returns.index.dayofweek < 5, other=np.nan)
    vol1 = qis.compute_ewm_vol(data=returns1, span=span, annualization_factor=260*24, init_value=init_value)
    vols = pd.concat([vol.rename('including weekends'),
                      vol1.rename('excluding weekends')], axis=1)
    return vols


class LocalTests(Enum):
    VOLS = 1


def run_local_test(local_test: LocalTests):
    """Run local tests for development and debugging purposes.

    These are integration tests that download real data and generate reports.
    Use for quick verification during development.
    """

    if local_test == LocalTests.VOLS:
        prices = fetch_hourly_data(ticker='BTC-USD')['close']
        vol = compute_vols(prices=prices)
        with sns.axes_style("darkgrid"):
            fig, ax = plt.subplots(1, 1, figsize=(10, 7))
            qis.plot_time_series(df=vol, var_format='{:,.2%}', ax=ax)

    plt.show()


if __name__ == '__main__':

    run_local_test(local_test=LocalTests.VOLS)



"""
illustrate interpolation of infrequent returns
"""
# packages
import pandas as pd
import numpy as np
import matplotlib.pyplot as plt
from enum import Enum
import qis as qis


class LocalTests(Enum):
    YF = 1
    BBG = 2


def run_local_test(local_test: LocalTests):
    """Run local tests for development and debugging purposes.

    These are integration tests that download real data and generate reports.
    Use for quick verification during development.
    """

    pd.set_option('display.max_rows', 500)
    pd.set_option('display.max_columns', 500)
    pd.set_option('display.width', 1000)

    if local_test == LocalTests.YF:
        import yfinance as yf
        pivot = 'SPY'
        asset = 'QQQ'
        tickers = [pivot, asset]
        prices = yf.download(tickers=tickers, start="2003-12-31", end=None, ignore_tz=True, auto_adjust=True)['Close']

    elif local_test == LocalTests.BBG:
        from bbg_fetch import fetch_field_timeseries_per_tickers
        pivot = 'SPTR Index'
        asset = 'XNDX Index'
        tickers = [pivot, asset]
        prices = fetch_field_timeseries_per_tickers(tickers=tickers, freq='B', field='PX_LAST').ffill()

    else:
        raise NotImplementedError

    is_log_returns = True
    infrequent_returns = qis.to_returns(prices[asset], is_log_returns=is_log_returns, freq='QE')
    pivot_returns = qis.to_returns(prices[pivot], is_log_returns=is_log_returns, freq='ME')
    i_backfill = qis.interpolate_infrequent_returns(infrequent_returns=infrequent_returns.dropna(), pivot_returns=pivot_returns,
                                                    is_to_log_returns=False)

    known_returns = qis.to_returns(prices[asset], is_log_returns=is_log_returns, freq='ME')
    returns = pd.concat([pivot_returns, known_returns.rename(f"{asset} actual"), i_backfill.rename(f"{asset} interpolated")], axis=1).dropna()
    if is_log_returns:
        returns = np.expm1(returns)
    print(f"means={np.nanmean(returns, axis=0)}, stdevs={np.nanstd(returns, axis=0)}")

    navs = qis.returns_to_nav(returns)
    qis.plot_time_series(navs)
    returns = pd.concat([returns.rolling(3).sum(), infrequent_returns.rename('Q')], axis=1)
    print(returns)

    fig = plt.figure(figsize=(12, 10), constrained_layout=True)
    gs = fig.add_gridspec(nrows=3, ncols=2, wspace=0.0, hspace=0.0)
    qis.plot_ra_perf_table(prices=navs, perf_params=qis.PerfParams(freq='ME'), title='Monthly Sampling', ax=fig.add_subplot(gs[0, :]))
    qis.plot_ra_perf_table(prices=navs, perf_params=qis.PerfParams(freq='QE'), title='Quarterly Sampling', ax=fig.add_subplot(gs[1, :]))

    qis.plot_returns_corr_table(prices=navs, freq='ME', title='Monthly Sampling', ax=fig.add_subplot(gs[2, 0]))
    qis.plot_returns_corr_table(prices=navs, freq='QE', title='Quarterly Sampling', ax=fig.add_subplot(gs[2, 1]))

    plt.show()


if __name__ == '__main__':

    run_local_test(local_test=LocalTests.YF)



"""
plot aggregate report on pandas data and plot time series of columns
each time series plot = data[column]
 # = # columns
"""
# packages
import pandas as pd
import numpy as np
import seaborn as sns
import matplotlib.pyplot as plt
from typing import Tuple, Optional, Dict, List
from enum import Enum

# qis
import qis as qis
import qis.utils.dates as da
import qis.utils.df_agg as dfa
import qis.utils.df_groups as dfg
import qis.utils.df_ops as dfo
from qis.perfstats.desc_table import DescTableType
from qis.plots import histogram as pdf
from qis.plots.reports.utils import get_summary_table_fig, set_x_date_freq, ReportType
from qis.plots.derived.data_timeseries import plot_data_timeseries


def econ_data_report(data: pd.DataFrame,
                     complimentary_datas: Dict[str, pd.DataFrame] = None,
                     group_data: pd.Series = None,
                     descriptive_data: pd.DataFrame = None,
                     inst_names: pd.Series = None,
                     feature_name: str = None,
                     report_type: ReportType = ReportType.SingleTimeSeries,
                     is_add_summary_table: bool = True,
                     file_name: str = 'econ_data',
                     var_format: Optional[str] = None,
                     is_price_data: bool = False,
                     desc_table_type: DescTableType = DescTableType.WITH_KURTOSIS,
                     x_min_max_quantiles: Tuple[Optional[float], Optional[float]] = (0.005, 0.995),
                     background: str = 'darkgrid',
                     **kwargs
                     ) -> List[plt.Figure]:
    """
    report for economic data
    """
    local_kwargs = {'fontsize': 6,
                    'linewidth': 0.5,
                    'weight': 'normal',
                    'framealpha': 0.75}
    qis.update_kwargs(local_kwargs, kwargs)

    figsize_table = (8.0, 2.9)  # table for 1 column
    figsize = (14, 10)

    if var_format is None:
        if is_price_data:
            var_format = '{:,.2f}'
        else:
            var_format = '{:.2%}'

    # get summary table
    price_start_dates = {}
    price_end_dates = {}
    insample_prices = {}
    for inst in data.columns:
        insample_price = dfo.drop_first_nan_data(df=data[inst]).dropna()

        if insample_price.empty:
            bbg_start = np.nan
            bbg_finish = np.nan
        else:
            bbg_start = insample_price.index[0]
            bbg_finish = insample_price.index[-1]

        price_start_dates.update({inst: bbg_start})
        price_end_dates.update({inst: bbg_finish})
        insample_prices.update({inst: insample_price})
    price_start_dates = pd.Series(price_start_dates).rename('StartDate')
    price_end_dates = pd.Series(price_end_dates).rename('EndDate')
    price_start_dates = pd.concat([price_start_dates, price_end_dates], axis=1)

    figs = []
    if is_add_summary_table:
        fig = get_summary_table_fig(data_start_dates=price_start_dates,
                                    descriptive_data=descriptive_data,
                                    figsize=figsize_table,
                                    **kwargs)
        figs.append(fig)

    group_means = None
    if group_data is not None:
        group_means = dfg.agg_df_by_groups(df=data,
                                           group_data=group_data,
                                           agg_func=dfa.nanmean,
                                           total_column=None)
    figure_num = 0
    for inst, insample_price in insample_prices.items():
        if np.all(insample_price.isnull()):  # skip for all nans
            continue
        figure_num += 1
        figure_caption = f"Figure {figure_num}. {inst}"
        if inst_names is not None:
            figure_caption += f", {inst_names[inst]}"

        if descriptive_data is not None:
            for k, v in descriptive_data.loc[inst].to_dict().items():
                figure_caption += f", {k}: {v}"

        local_kwargs = set_x_date_freq(data=insample_price, kwargs=kwargs)

        # change names
        if feature_name is None:
            inst_name = f"{inst} "
        else:
            insample_price = insample_price.rename(feature_name)
            inst_name = ""

        if complimentary_datas is not None:
            for name, data in complimentary_datas.items():
                if inst in data.columns:
                    comp_name = f"{inst_name}{name}"
                    insample_price = pd.concat([insample_price, data[inst].rename(comp_name)], axis=1)

        if group_means is not None:
            inst_group = group_data[inst]
            insample_price = pd.concat([insample_price, group_means[inst_group].rename(f"{inst_group} ac avg")], axis=1)

        if report_type == ReportType.SingleTimeSeries:
            with sns.axes_style(background):
                fig, ax = plt.subplots(1, 1, figsize=figsize, tight_layout=True)
                plot_data_timeseries(data=insample_price,
                                     ax=ax,
                                     title=f"In-sample period = ",
                                     is_price_data=is_price_data,
                                     var_format=var_format,
                                     **local_kwargs)

        elif report_type == ReportType.SingleTimeSeriesWithPDF:
            fig, axs = plt.subplots(1, 2, figsize=figsize, tight_layout=True)
            plot_data_timeseries(data=insample_price,
                                 ax=axs[0],
                                 title=f"In-sample period = ",
                                 is_price_data=is_price_data,
                                 var_format=var_format,
                                 **local_kwargs)

            pdf.plot_histogram(df=insample_price,
                               xvar_format=var_format,
                               x_min_max_quantiles=x_min_max_quantiles,
                               add_last_value=False,
                               desc_table_type=desc_table_type,
                               title=f"Distribution: {da.get_time_period(df=data).to_str()}",
                               bbox_to_anchor=None,
                               ax=axs[1],
                               **kwargs)

        else:
            raise TypeError(f"report_type={report_type} is not implemented")

        qis.set_suptitle(fig, title=f"{file_name}_{figure_num}_{inst}")
        figs.append(fig)

    return figs


class LocalTests(Enum):
    UPDATE_DATA = 1
    RUN_REPORT = 2


def run_local_test(local_test: LocalTests):
    """Run local tests for development and debugging purposes.

    These are integration tests that download real data and generate reports.
    Use for quick verification during development.
    """

    import matplotlib.pyplot as plt
    if local_test == LocalTests.UPDATE_DATA:
        from bbg_fetch import fetch_field_timeseries_per_tickers
        vix_tickers = ['VIX1D Index', 'VIX9D Index', 'VIX Index', 'VIX3M Index', 'VIX6M Index', 'VIX1Y Index']
        vols = 0.01 * fetch_field_timeseries_per_tickers(tickers=vix_tickers, field='PX_LAST', CshAdjNormal=True)
        print(vols)
        qis.save_df_to_csv(df=vols, file_name='vix_indices')

    elif local_test == LocalTests.RUN_REPORT:
        df = qis.load_df_from_csv(file_name='vix_indices')
        figs = econ_data_report(data=df)
        qis.save_figs_to_pdf(figs, file_name='vix_indices')

    plt.show()


if __name__ == '__main__':

    run_local_test(local_test=LocalTests.RUN_REPORT)



"""
plot descriptive table
"""
# packages
import pandas as pd
import matplotlib.pyplot as plt
from typing import Union, Optional
from enum import Enum

# qis
from qis.plots.table import plot_df_table
from qis.perfstats.desc_table import compute_desc_table, DescTableType


def plot_desc_table(df: Union[pd.DataFrame, pd.Series],
                    desc_table_type: DescTableType = DescTableType.SHORT,
                    var_format: str = '{:.2f}',
                    annualize_vol: bool = False,
                    is_add_tstat: bool = False,
                    norm_variable_display_type: str = '{:.1f}',  # for t-stsat
                    ax: plt.Subplot = None,
                    **kwargs
                    ) -> Optional[plt.Figure]:
    """
    data corresponds to matrix of returns with index = time and columns = tickers
    data can contain nan in columns
    output is index = tickers, columns = descriptive data
    data converted to str
    """
    table_data = compute_desc_table(df=df,
                                    desc_table_type=desc_table_type,
                                    var_format=var_format,
                                    annualize_vol=annualize_vol,
                                    is_add_tstat=is_add_tstat,
                                    norm_variable_display_type=norm_variable_display_type,
                                    **kwargs)

    fig = plot_df_table(df=table_data,
                        ax=ax,
                        **kwargs)
    return fig


class LocalTests(Enum):
    TABLE = 1


def run_local_test(local_test: LocalTests):
    """Run local tests for development and debugging purposes.

    These are integration tests that download real data and generate reports.
    Use for quick verification during development.
    """

    from qis.test_data import load_etf_data
    returns = load_etf_data().dropna().asfreq('QE').pct_change()

    if local_test == LocalTests.TABLE:
        plot_desc_table(df=returns,
                        desc_table_type=DescTableType.EXTENSIVE,
                        var_format='{:.2f}',
                        annualize_vol=True,
                        is_add_tstat=False)

    plt.show()


if __name__ == '__main__':

    run_local_test(local_test=LocalTests.TABLE)



"""
implementation of bootsrap methods in numba
"""

import numpy as np
import pandas as pd
from enum import Enum
from numba import njit
from numba.typed import List
from typing import Union, Tuple, Dict
from statsmodels.tsa.ar_model import AutoReg

# qis
import qis.utils.np_ops as npo
import qis.perfstats.returns as ret


class BootsrapType(Enum):
    IID = 1
    STATIONARY = 2


class BootsrapOutput(Enum):
    SERIES_TO_DF = 1
    DF_TO_LIST_ARRAYS = 2


@njit
def set_seed(value):
    """
    set seed for numba space
    """
    np.random.seed(value)


@njit
def bootstrap_indices_iid(num_data_index: int,
                          num_samples: int = 10,
                          index_length: int = 1000,
                          seed: int = 1
                          ) -> np.ndarray:
    """
    generate iid bootstrap indices
    """
    np.random.seed(seed)
    set_seed(seed)
    bootstrapped_indices = np.zeros((index_length, num_samples), dtype=np.int64)
    for idx in np.arange(num_samples):
        previous_index, next_index = 0, 0
        while next_index < index_length-1:
            ran_indices = np.random.randint(low=0, high=num_data_index, size=num_data_index)
            next_index = np.minimum(previous_index+num_data_index, index_length-1)
            required_fill = next_index - previous_index
            bootstrapped_indices[previous_index:next_index, idx] = ran_indices[:required_fill]
            previous_index = next_index
    return bootstrapped_indices


@njit
def bootstrap_indices_stationary(num_data_index: int,
                                 num_samples: int = 10,
                                 num_bootstrap_index: int = 1000,
                                 block_size: int = 30,
                                 seed: int = 1
                                 ) -> np.ndarray:
    """
    generate stationary bootstrap indices
    """
    np.random.seed(seed)
    set_seed(seed)
    bootstrapped_indices = np.zeros((num_bootstrap_index, num_samples), dtype=np.int64)
    for idx in np.arange(num_samples):
        previous_index, next_index = 0, 0
        while next_index < num_bootstrap_index-1:
            start_index = np.random.randint(low=0, high=num_data_index) # random start of a block
            random_block_size = np.random.geometric(1.0/block_size)  # random block size, default size is 1
            end_index_in_data = np.minimum(start_index+random_block_size, num_data_index)  # end of sample
            proposed_fill = end_index_in_data-start_index  # how much we can increase based on index in data
            next_index = np.minimum(previous_index + proposed_fill, num_bootstrap_index)  # index for backfill indices
            implemented_fill = next_index-previous_index  # how much we will increase based on index in bootstrap indices
            bootstrapped_indices[previous_index:next_index, idx] = np.arange(start_index, start_index+implemented_fill)
            previous_index = next_index
    return bootstrapped_indices


@njit
def get_bootsrtap_data_list(data_np: np.ndarray,
                            bootstrapped_indices: np.ndarray
                            ) -> List:
    """
    map indices to data
    """
    bootstrap_sample = List()
    for index in np.transpose(bootstrapped_indices):
        bootstrapped_data = data_np[index, :]
        bootstrap_sample.append(bootstrapped_data)
    return bootstrap_sample


@njit
def get_bootsrtap_ar_data_list(residuals: np.ndarray,
                               intercept: np.ndarray,
                               beta: np.ndarray,
                               data0: np.ndarray,
                               bootstrapped_indices: np.ndarray,
                               is_positive: bool = True
                               ) -> List:
    """
    map indices to ar data
    """
    bootstrap_sample = List()
    for index in np.transpose(bootstrapped_indices):
        bootstrapped_resids = residuals[index, :]
        bootstrapped_data = np.zeros_like(bootstrapped_resids)
        y0 = data0*np.ones_like(beta)  # nb important for numba so it can map beta*y0
        for idx, resid in enumerate(bootstrapped_resids):
            y0 = intercept + beta*y0 + resid
            if is_positive:
                y0 = np.where(np.greater(y0, 0.0), y0, np.nanquantile(y0, 0.25))
            bootstrapped_data[idx] = y0

        bootstrap_sample.append(bootstrapped_data)
    return bootstrap_sample


def compute_ar_residuals(data: Union[pd.Series, pd.DataFrame]
                         ) -> Tuple[np.ndarray, np.ndarray, np.ndarray]:
    if isinstance(data, pd.Series):
        data = data.to_frame()
    intercept, beta = np.zeros(len(data.columns)), np.zeros(len(data.columns))
    residuals = np.zeros((len(data.index)-1, len(data.columns)))
    for idx, column in enumerate(data.columns):
        ar_model = AutoReg(data[column].dropna(), lags=1).fit()
        intercept[idx], beta[idx] = ar_model.params[0], ar_model.params[1]
        data_np = data[column].to_numpy()
        y, y_1 = data_np[1:], data_np[:-1]
        residuals[:, idx] = y - (ar_model.params[1] * y_1 + ar_model.params[0])
    return residuals, intercept, beta


def generate_bootstrapped_indices(num_data_index: int,
                                  bootsrap_type: BootsrapType = BootsrapType.IID,
                                  num_samples: int = 10,
                                  index_length: int = 1000,
                                  block_size: int = 30,
                                  seed: int = 1
                                  ) -> np.ndarray:
    """
    wrapper for numba function
    """
    if bootsrap_type == BootsrapType.IID:
        bootstrapped_indices = bootstrap_indices_iid(num_data_index=num_data_index,
                                                     num_samples=num_samples,
                                                     index_length=index_length,
                                                     seed=seed)
    elif bootsrap_type == BootsrapType.STATIONARY:
        bootstrapped_indices = bootstrap_indices_stationary(num_data_index=num_data_index,
                                                            num_samples=num_samples,
                                                            num_bootstrap_index=index_length,
                                                            block_size=block_size,
                                                            seed=seed)
    else:
        raise ValueError(f"not implemented")

    return bootstrapped_indices


def bootstrap_data(data: Union[pd.Series, pd.DataFrame],
                   bootsrap_type: BootsrapType = BootsrapType.STATIONARY,
                   bootsrap_output: BootsrapOutput = BootsrapOutput.DF_TO_LIST_ARRAYS,
                   num_samples: int = 10,
                   index_length: int = 1000,
                   block_size: int = 30,
                   seed: int = 1,
                   bootstrapped_indices: np.ndarray = None
                   ) -> Union[List, pd.DataFrame]:
    """
    core data bootsrap
    """
    if bootstrapped_indices is None:
        bootstrapped_indices = generate_bootstrapped_indices(num_data_index=len(data.index),
                                                             bootsrap_type=bootsrap_type,
                                                             num_samples=num_samples,
                                                             index_length=index_length,
                                                             block_size=block_size,
                                                             seed=seed)

    if bootsrap_output == BootsrapOutput.DF_TO_LIST_ARRAYS:
        bootstrap_sample = get_bootsrtap_data_list(data_np=data.to_numpy(),
                                                   bootstrapped_indices=bootstrapped_indices)

    elif bootsrap_output == BootsrapOutput.SERIES_TO_DF:
        if not isinstance(data, pd.Series):
            raise ValueError(f"data must be series")

        bootstrap_sample = get_bootsrtap_data_list(data_np=npo.np_array_to_matrix(a=data.to_numpy(), ncols=1),
                                                   bootstrapped_indices=bootstrapped_indices)
        data = []
        for idx, sample in enumerate(bootstrap_sample):
            data.append(pd.DataFrame(sample, columns=[f"path_{idx+1}"]))
        bootstrap_sample = pd.concat(data, axis=1)

    else:
        raise ValueError(f"not implemented")

    return bootstrap_sample


def bootstrap_ar_process(data: Union[pd.Series, pd.DataFrame],
                         bootsrap_type: BootsrapType = BootsrapType.STATIONARY,
                         bootsrap_output: BootsrapOutput = BootsrapOutput.DF_TO_LIST_ARRAYS,
                         num_samples: int = 10,
                         index_length: int = 1000,
                         block_size: int = 30,
                         seed: int = 1,
                         bootstrapped_indices: np.ndarray = None
                         ) -> Union[List, pd.DataFrame]:

    residuals, intercept, beta = compute_ar_residuals(data=data)
    if bootstrapped_indices is None:
        bootstrapped_indices = generate_bootstrapped_indices(num_data_index=len(data.index),
                                                             bootsrap_type=bootsrap_type,
                                                             num_samples=num_samples,
                                                             index_length=index_length,
                                                             block_size=block_size,
                                                             seed=seed)

    if bootsrap_output == BootsrapOutput.DF_TO_LIST_ARRAYS:
        bootstrap_sample = get_bootsrtap_ar_data_list(residuals=residuals,
                                                      intercept=intercept,
                                                      beta=beta,
                                                      data0=np.nanmean(data, axis=0),
                                                      bootstrapped_indices=bootstrapped_indices)

    elif bootsrap_output == BootsrapOutput.SERIES_TO_DF:
        if not isinstance(data, pd.Series):
            raise ValueError(f"data must be series")

        bootstrap_sample = get_bootsrtap_ar_data_list(residuals=residuals,
                                                      intercept=intercept,
                                                      beta=beta,
                                                      data0=np.array(np.nanmean(data)),
                                                      bootstrapped_indices=bootstrapped_indices)

        data = []
        for idx, sample in enumerate(bootstrap_sample):
            data.append(pd.DataFrame(sample, columns=[f"path_{idx+1}"]))
        bootstrap_sample = pd.concat(data, axis=1)

    else:
        raise ValueError(f"not implemented")

    return bootstrap_sample


def bootstrap_price_data(prices: Union[pd.Series, pd.DataFrame],
                         bootsrap_type: BootsrapType = BootsrapType.STATIONARY,
                         bootsrap_output: BootsrapOutput = BootsrapOutput.DF_TO_LIST_ARRAYS,
                         num_samples: int = 10,
                         index_length: int = 1000,
                         block_size: int = 20,
                         is_log_returns: bool = False,
                         seed: int = 1,
                         bootstrapped_indices: np.ndarray = None,
                         init_to_end: bool = True
                         ) -> Union[List[np.ndarray], pd.DataFrame]:
    """
    bootstrap price data
    for pd.Dataframe use bootsrap_output = BootsrapOutput.DF_TO_LIST_ARRAYS to get list of nd.arrays
    block_size = 1 corresponds to iid sampling
    """
    returns = ret.to_returns(prices=prices, is_log_returns=is_log_returns, drop_first=True)

    bootstrap_returns = bootstrap_data(data=returns,
                                       bootsrap_type=bootsrap_type,
                                       bootsrap_output=bootsrap_output,
                                       num_samples=num_samples,
                                       index_length=index_length,
                                       block_size=block_size,
                                       seed=seed,
                                       bootstrapped_indices=bootstrapped_indices)

    if bootsrap_output == BootsrapOutput.DF_TO_LIST_ARRAYS:
        if init_to_end:
            init_value = prices.iloc[-1, :].to_numpy()
        else:
            init_value = prices.iloc[0, :].to_numpy()

        bootstrap_sample = List()
        for returns in bootstrap_returns:
            if is_log_returns:
                bootstrap_sample.append(ret.log_returns_to_nav(log_returns=returns, init_value=init_value))
            else:
                bootstrap_sample.append(ret.returns_to_nav(returns=returns, init_value=init_value))

    elif bootsrap_output == BootsrapOutput.SERIES_TO_DF:
        if init_to_end:
            init_value = prices[-1]*np.ones(num_samples)
        else:
            init_value = prices[0]*np.ones(num_samples)

        if is_log_returns:
            bootstrap_sample = ret.log_returns_to_nav(log_returns=bootstrap_returns, init_value=init_value)
        else:
            bootstrap_sample = ret.returns_to_nav(returns=bootstrap_returns, init_value=init_value)
    else:
        raise ValueError(f"not implemented")

    return bootstrap_sample


def bootstrap_price_fundamental_data(price_datas: Dict[str, Union[pd.Series, pd.DataFrame]],
                                     fundamental_datas: Dict[str, Union[pd.Series, pd.DataFrame]],
                                     bootsrap_type: BootsrapType = BootsrapType.STATIONARY,
                                     bootsrap_output: BootsrapOutput = BootsrapOutput.DF_TO_LIST_ARRAYS,
                                     num_samples: int = 10,
                                     index_length: int = 1000,
                                     block_size: int = 30,
                                     is_log_returns: bool = False,
                                     seed: int = 1,
                                     is_price_weighted_fundamentals: bool = False  # multiply by price_datas[0] price bootstrap
                                     ) -> Tuple[Dict[str, Union[pd.DataFrame, List]], Dict[str, Union[pd.DataFrame, List]]]:

    """
    price data in the first element must  be aligned with all fundamentals
    """
    prices = price_datas[list(price_datas.keys())[0]]
    for key, fundamental_data in fundamental_datas.items():
        assert all(prices.index == fundamental_data.index)
        if isinstance(prices, pd.DataFrame) and isinstance(fundamental_data, pd.DataFrame):
            assert all(prices.columns == fundamental_data.columns)
        elif isinstance(prices, pd.Series) and isinstance(fundamental_data, pd.Series):
            pass
        else:
            raise ValueError(f" data types not aligned")

    # very important to reduce lenth for returns and ar-1 bootstrap
    bootstrapped_indices = generate_bootstrapped_indices(num_data_index=len(prices.index)-1,
                                                         bootsrap_type=bootsrap_type,
                                                         num_samples=num_samples,
                                                         index_length=index_length,
                                                         block_size=block_size,
                                                         seed=seed)

    bootstrap_prices = {}
    for key, prices in price_datas.items():
        bootstrap_prices[key] = bootstrap_price_data(prices=prices,
                                                     bootsrap_type=bootsrap_type,
                                                     bootsrap_output=bootsrap_output,
                                                     num_samples=num_samples,
                                                     index_length=index_length,
                                                     block_size=block_size,
                                                     is_log_returns=is_log_returns,
                                                     seed=seed,
                                                     bootstrapped_indices=bootstrapped_indices)

    bootstrap_fundamentals = {}
    for key, fundamental_data in fundamental_datas.items():
        fundamentals = bootstrap_ar_process(data=fundamental_data,
                                            bootsrap_type=bootsrap_type,
                                            bootsrap_output=bootsrap_output,
                                            num_samples=num_samples,
                                            index_length=index_length,
                                            block_size=block_size,
                                            seed=seed,
                                            bootstrapped_indices=bootstrapped_indices)

        if is_price_weighted_fundamentals:
            prices = bootstrap_prices[list(bootstrap_prices.keys())[0]]
            for idx, (price, fund) in enumerate(zip(prices, fundamentals)):
                fundamentals[idx] = price*fund

        bootstrap_fundamentals[key] = fundamentals

    return bootstrap_prices, bootstrap_fundamentals


class LocalTests(Enum):
    DATA_LIST = 1
    DATA_SERIES = 2
    BTC = 3


def run_local_test(local_test: LocalTests):
    """Run local tests for development and debugging purposes.

    These are integration tests that download real data and generate reports.
    Use for quick verification during development.
    """

    # simulate t*n data
    dates = pd.date_range(start='12/31/2018', end='12/31/2019', freq='B')
    n = 3
    mean = [-2.0, -1.0, 0.0]
    data = pd.DataFrame(data=np.random.normal(mean, 1.0, (len(dates), n)),
                        index=dates,
                        columns=['x'+str(m+1) for m in range(n)])
    print(data)

    if local_test == LocalTests.DATA_LIST:
        bt_lists = bootstrap_data(data=data,
                                  bootsrap_type=BootsrapType.STATIONARY,
                                  bootsrap_output=BootsrapOutput.DF_TO_LIST_ARRAYS)
        for this in bt_lists:
            print(this)

    elif local_test == LocalTests.DATA_SERIES:
        bt_data = bootstrap_data(data=data.iloc[:, 0],
                                 bootsrap_type=BootsrapType.STATIONARY,
                                 bootsrap_output=BootsrapOutput.SERIES_TO_DF)
        print(bt_data)

    elif local_test == LocalTests.BTC:
        from legacy.crypto.data.apis.coinmetric_ import load_btc_price
        pivot_prices = load_btc_price().dropna()
        bt_data = bootstrap_price_data(prices=pivot_prices,
                                       bootsrap_type=BootsrapType.STATIONARY,
                                       bootsrap_output=BootsrapOutput.DF_TO_LIST_ARRAYS)
        print(bt_data)


if __name__ == '__main__':

    run_local_test(local_test=LocalTests.BTC)



"""
example of creating 60/40 equity with and without BTC
"""
import pandas as pd
import matplotlib.pyplot as plt
import seaborn as sns
import qis as qis
import yfinance as yf

btc_weight = 0.02
tickers_weights_wo = {'SPY': 0.6, 'IEF': 0.4, 'BTC-USD': 0.0}
tickers_weights_with = {'SPY': 0.6-0.5*btc_weight, 'IEF': 0.4-0.5*btc_weight, 'BTC-USD': btc_weight}


tickers = list(tickers_weights_wo.keys())
prices = yf.download(tickers, start="2003-12-31", end=None, ignore_tz=True, auto_adjust=True)['Close'][tickers].asfreq('B', method='ffill').dropna()
prices = prices.loc['31Dec2018':, :]

balanced_60_40_wo = qis.backtest_model_portfolio(prices=prices, weights=tickers_weights_wo, rebalancing_freq='QE',
                                                 ticker='100% Balanced').get_portfolio_nav()
balanced_60_40b = qis.backtest_model_portfolio(prices=prices, weights=tickers_weights_with, rebalancing_freq='QE',
                                               ticker=f"{1.0-btc_weight:0.0%} Balanced / {btc_weight:0.0%} BTC").get_portfolio_nav()
navs = pd.concat([balanced_60_40_wo, balanced_60_40b], axis=1)

perf_params = qis.PerfParams(freq='ME')
with sns.axes_style("darkgrid"):
    fig = plt.figure(figsize=(15, 12), constrained_layout=True)
    gs = fig.add_gridspec(nrows=7, ncols=1, wspace=0.0, hspace=0.0)
    axs = [fig.add_subplot(gs[1:4, 0]), fig.add_subplot(gs[4:, 0])]
    qis.plot_prices_with_dd(prices=navs, perf_params=perf_params, axs=axs)
    qis.plot_ra_perf_table_benchmark(prices=navs,
                                     benchmark=balanced_60_40_wo.name,
                                     perf_params=perf_params,
                                     title='Risk-adjusted Performance Table',
                                     digits_to_show=1,
                                     ax=fig.add_subplot(gs[0, 0]))

plt.show()
